# -*- coding: utf-8 -*-
from jax import numpy as jnp
from jax import Array, jit, vmap
from jax.typing import ArrayLike
from jax.scipy.special import erfc


def point_heat_source(p_source: Array, p: Array, time: Array | float, alpha: float, r_min: float = 0.) -> Array | float:
    """Point heat source solution.

    Parameters
    ----------
    p_source : array or float
        (N, 3,) array of the positions of the point heat sources along
        the trajectory.
    p : array
        (M, 3,) array of the positions at which the point heat source
        solution is evaluated.
    time : array or float
        (K,) array of times (in seconds).
    alpha : float
        Ground thermal diffusivity (in m^2/s).
    r_min : float, default: ``0.``
        Minimum distance (in meters) between point heat sources and
        positions `p`.

    Returns
    -------
    array or float
        (K, M, N,) array of values of the point heat source solution.
        For each of the parameters `p_source`, `p` and `time`, the
        corresponding axis is removed if the parameter is supplied as
        a ``float`` (for `time`) or a 1d array (for `p_source` and
        `p`).

    """
    if len(jnp.shape(time)) > 0:
        return vmap(
            point_heat_source,
            in_axes=(None, None, -1, None, None, None)
        )(p_source, p, time, J, alpha, r_min)
    if len(jnp.shape(p)) > 1:
        return vmap(
            point_heat_source,
            in_axes=(None, -2, None, None, None, None)
        )(p_source, p, time, J, alpha, r_min)
    if len(jnp.shape(p_source)) > 1:
        return vmap(
            point_heat_source,
            in_axes=(-2, None, None, 0, None, None)
        )(p_source, p, time, J, alpha, r_min)
    # Distance to the real point (p)
    r = jnp.sqrt(((p_source - p)**2).sum() + r_min**2)
    return _point_heat_source(r, time, alpha)


@jit
def _point_heat_source(r: float, time: float, alpha: float) -> float:
    """Point heat source solution.

    Parameters
    ----------
    r : float
        Distance between the point source and the evaluation point (in
        meters).
    time : float
        Time (in seconds).
    alpha : float
        Ground thermal diffusivity (in m^2/s).

    Returns
    -------
    float
        Point heat source solution.

    """
    # Point heat source solution
    h = 0.5 * erfc(r / jnp.sqrt(4 * alpha * time)) / r
    return h



# -*- coding: utf-8 -*-
from collections.abc import Callable
from functools import partial
from itertools import cycle
from time import perf_counter
from typing import Tuple

from jax import numpy as jnp
from jax import Array, debug, jit, vmap
from jax.lax import cond, fori_loop
from jax.typing import ArrayLike
import numpy as np

from ..borefield.network import Network
from .load_aggregation import LoadAggregation


class Simulation:
    """Simulation.

    Parameters
    ----------
    borefield: network
        The borefield.
    dt : float
        Time step (in seconds).
    tmax : float
        Maximum time of the simulation (in seconds).
    T0 : float or callable
        Undisturbed ground temperature (in degree Celcius), or callable
        that takes the time (in seconds) and a 1d array of positions ``z``
        (in meters, negative) as inputs and returns a 1d array of
        undisturbed ground temperatures (in degree Celsius).
    cp_f : float
        Fluid specific isobaric heat capacity (in J/kg-K).
    alpha : float
        Ground thermal diffusivity (in m^2/s)
    k_s : float
        Ground thermal conductivity (in W/m-K).
    cells_per_level : int, default: ``5``
        Number of cells per aggregation level.
    p : array_like or None, default: ``None``
        (`n_points`, 3,) array of positions to evaluate the ground
        temperature. If `p` is ``None``, the ground temperature is not
        evaluated.
    store_node_values : bool, default: ``False``
        Set to True to store the borehole wall temperature and heat
        extraction rates at the nodes.

    Attributes
    ----------
    n_times : int
        Number of time steps.
    q : array
        The heat extraction rate at the nodes (in W/m).
        Only if `store_node_values` if ``True``.
    Q : array
        The total heat extraction rate (in watts).
    T_b : array
        The borehole wall temperature at the nodes (in degree Celsius).
        Only if `store_node_values` if ``True``.
    T_f_in : array
        The inlet fluid temperature (in degree Celsius).
    T_f_out : array
        The outlet fluid temperature (in degree Celsius).
    T : array
        The ground temperature (in degree Celsius) at positions `p`.
    m_flow : float
        Total fluid mass flow rate (in kg/s).

    """

    def __init__(self, borefield: Network, cp_f: float, dt: float, tmax: float, T0: float | Callable[[float, Array], Array | float], alpha: float, k_s: float, cells_per_level: int = 5, p: ArrayLike | None = None, store_node_values: bool = False):
        # Runtime type validation
        if not isinstance(p, ArrayLike) and p is not None:
            raise TypeError(f"Expected arraylike or None input; got {p}")
        # Convert input to jax.Array
        if p is not None:
            p = jnp.asarray(p)

        # --- Class atributes ---
        # Parameters
        self.borefield = borefield
        self.cp_f = cp_f
        self.dt = dt
        self.tmax = tmax
        self.n_nodes = self.borefield.n_boreholes * self.borefield.n_nodes
        self.T0 = T0
        self.alpha = alpha
        self.k_s = k_s
        self.cells_per_level = cells_per_level
        self.p = p
        self.store_node_values = store_node_values
        # Additional attributes
        self.n_times = int(tmax // dt)
        if p is None:
            self.T = None
            self.n_points = 0
        else:
            self.n_points = p.shape[0]

        # --- Initialization ---
        self.loadAgg = LoadAggregation(
            borefield, dt, tmax, alpha, cells_per_level=cells_per_level, p=p)
        self.initialize_system_of_equations()
        # Convert `T0` to callable
        if not callable(T0):
            def undisturbed_ground_temperature(time: float, z: Array | float) -> float:
                return T0
            self.undisturbed_ground_temperature = undisturbed_ground_temperature
        else:
            self.undisturbed_ground_temperature = T0

    def initialize_system_of_equations(self):
        """Initialize the system of equations.

        """
        N = self.n_nodes
        # Thermal response factors
        self.h_to_self = self.loadAgg.h_to_self[0] / (2 * jnp.pi * self.k_s)
        # Initialize system of equations
        self.A = jnp.block(
            [[-jnp.eye(N), jnp.zeros((N, 1))],
             [self.borefield.w.flatten(), jnp.zeros((1, 1))]]
            )
        self.B = jnp.zeros(N + 1)

    def update_system_of_equations(self, m_flow: float | Array, Q: float, T0: Array) -> Tuple[Array, Array]:
        """Update the system of equations.

        Parameters
        ----------
        m_flow : float or array
            Total fluid mass flow rate (in kg/s), or (`n_boreholes`,)
            array of fluid mass flow rate per borehole.
        Q : float
            Total heat extraction rate (in watts).
        T0 : array
            Borehole wall temperature at nodes (in degree Celsius)
            assuming zero heat extraction rate.

        Returns
        -------
        A : array
            Linear system of equations.
        B : array
            Right-hand side of the linear system of equations.

        """
        N = self.n_nodes
        # Borehole heat transfer rate coefficients for current fluid mass
        # flow rate `m_flow`
        self.g_in, self.g_b = self.borefield.g_to_self(m_flow, self.cp_f)
        # Update system of equation for the current borehole wall
        # temperature `T0` at nodes
        A = self.A.at[:N, :N].set(-(jnp.eye(N) + jnp.einsum('iml,iljn->imjn', self.g_b, self.h_to_self).reshape((N, N))))
        A = A.at[:N, -1].set(self.g_in.flatten())
        B = self.B.at[:N].set(-vmap(jnp.dot, in_axes=(0, 0), out_axes=0)(self.g_b, T0.reshape((self.borefield.n_boreholes, -1))).flatten())
        # Apply total heat extraction rate `Q`
        B = B.at[-1].set(Q)
        return A, B

    def simulate(self, Q: ArrayLike, f_m_flow: Callable[[float], float | Array], m_flow_small: float = 0.01, disp: bool = True, print_every: int = 100):
        """Simulate the borefield.

        Parameters
        ----------
        Q : array_like
            Total heat extraction rate (in watts) at each time step. If
            the array is shorter than the simulation, it is repeated to
            cover the length of the simulation.
        f_m_flow : callable
            Function that takes the total heat extraction rate (in watts)
            as an input and returns the total fluid mas flow rate
            (in kg/s) or an array of fluid mass flow rate per borehole.
        m_flow_small : float, default: ``0.01``
            The minimum fluid mass flow rate (in kg/s). If `f_m_flow`
            returns a smaller total value, the fluid mass flow rate and
            the heat extraction rate are set to zero. If the total value
            is above `m_flow_small`, the minimum fluid mas flow rate per
            borehole is set to `m_flow_small`.
        disp : bool, default: ``True``
            Set to ``True`` to print simulation progression messages.
        print_every : int, default: ``100``
            Period at which simulation progression messages are printed.

        """
        tic = perf_counter()
        next_k = print_every
        if disp:
            print('Simulation start.')
        self.loadAgg.reset_history()

        # Initialize arrays
        Q = jnp.asarray(Q)
        n_cycles = np.ceil(self.n_times / len(Q)).astype(int)
        self.T_f_in = jnp.zeros(self.n_times)
        self.T_f_out = jnp.zeros(self.n_times)
        self.Q = jnp.tile(Q, n_cycles)[:self.n_times]
        self.m_flow = jnp.zeros(self.n_times)
        if self.store_node_values:
            self.q = jnp.zeros((self.n_times, self.borefield.n_boreholes, self.borefield.n_nodes))
            self.T_b = jnp.zeros((self.n_times, self.borefield.n_boreholes, self.borefield.n_nodes))
        else:
            self.q = None
            self.T_b = None
        if self.p is not None:
            self.T = jnp.zeros((self.n_times, self.n_points))
        else:
            self.T = None

        def simulate_step(
            i: int,
            val: Tuple[Array, Array, Array, Array, Array, Array | None, Array | None, Array]
        ) -> Tuple[Array, Array, Array, Array, Array, Array | None, Array | None, Array]:
            """Single simulation step."""
            # Unpack values
            m_flow, Q, T_b, T_f_in, T_f_out, q, T, q_history = val
            # Time step
            q_history = self.loadAgg._next_time_step(
                self.loadAgg.A, q_history)
            time = (i + 1) * self.dt
            current_Q = self.Q[i]
            # Fluid mass flow rate
            _m_flow = f_m_flow(current_Q)
            m_flow_network = jnp.sum(_m_flow)
            m_flow = m_flow.at[i].set(m_flow_network)
            # Temporal and spatial superposition of past loads
            T0 = (
                self.undisturbed_ground_temperature(
                    time, self.borefield.p[..., 2]
                )
                - self.loadAgg._temperature(
                    self.loadAgg.h_to_self,
                    q_history
                ) / (2 * jnp.pi * self.k_s)
            )
            # Build and solve system of equations
            _Q, _q, _T_b, _T_f_in, _T_f_out = self._simulate_step(
                _m_flow,
                current_Q,
                T0,
                m_flow_small
            )
            # Apply latest heat extraction rates
            q_history = self.loadAgg._current_load(q_history, _q)
            # Store results
            Q = Q.at[i].set(_Q)
            T_f_in = T_f_in.at[i].set(_T_f_in)
            T_f_out = T_f_out.at[i].set(_T_f_out)
            if self.store_node_values:
                q = q.at[i].set(_q)
                T_b = T_b.at[i].set(_T_b)
            # Evaluate ground temperatures
            if self.p is not None:
                T = T.at[i].set(
                    self.undisturbed_ground_temperature(
                        time, self.p[:, 2]
                    )
                    - self.loadAgg._temperature_to_point(
                        self.loadAgg.h_to_point,
                        q_history
                    ) / (2 * jnp.pi * self.k_s)
                )
            if disp:
                toc = perf_counter()
                cond(
                    (i + 1) % print_every == 0,
                    lambda _: debug.print(
                        'Completed {i} of {n_times} time steps. ',
                        i=i+1, n_times=self.n_times
                    ),
                    lambda _: None,
                    None
                )
            return m_flow, Q, T_b, T_f_in, T_f_out, q, T, q_history

        # Pack variables
        val = self.m_flow, self.Q, self.T_b, self.T_f_in, self.T_f_out, self.q, self.T, self.loadAgg.q
        # Run simulation
        self.m_flow, self.Q, self.T_b, self.T_f_in, self.T_f_out, self.q, self.T, _ = fori_loop(
            0, self.n_times, simulate_step, val, unroll=False)

        if disp:
            toc = perf_counter()
            debug.print(
                'Simulation end. Elapsed time: {clock:.2f} seconds.',
                clock=toc-tic
            )

    @partial(jit, static_argnames=['self'])
    def _simulate_step(self, m_flow: float | Array, Q: float, T0: Array, m_flow_small: float) -> Tuple[float, Array, Array, float, float]:
        """Solve a single time step.

        Parameters
        ----------
        m_flow : float or array
            Total fluid mass flow rate (in kg/s), or array of fluid mass
            flow rate per borehole.
        Q : float
            The total heat extraction rate (in watts).
        T0 : array
            Borehole wall temperature at nodes assuming zero heat
            extraction rate (in degree Celsius).
        m_flow_small : float
            The minimum fluid mass flow rate (in kg/s). If `f_m_flow`
            returns a smaller total value, the fluid mass flow rate and
            the heat extraction rate are set to zero. If the total value
            is above `m_flow_small`, the minimum fluid mas flow rate per
            borehole is set to `m_flow_small`.

        Returns
        -------
        Q : float
            The total heat extraction rate (in watts).
        q : array
            The heat extraction rate at the nodes (in W/m).
        T_b : array
            The borehole wall temperature at the nodes (in degree
            Celsius).
        T_f_in : float
            The inlet fluid temperature (in degree Celsius).

        """
        m_flow_network = jnp.sum(m_flow)
        Q, q, T_b, T_f_in, T_f_out = cond(
            m_flow_network > m_flow_small,
            self._simulate_step_flow,
            self._simulate_step_no_flow,
            m_flow,
            Q,
            T0,
            m_flow_small
        )
        return Q, q, T_b, T_f_in, T_f_out

    @partial(jit, static_argnames=['self'])
    def _simulate_step_flow(self, m_flow: float | Array, Q: float, T0: Array, m_flow_small: float) -> Tuple[float, Array, Array, float, float]:
        """Solve a single time step.

        Parameters
        ----------
        m_flow : float or array
            Total fluid mass flow rate (in kg/s), or array of fluid mass
            flow rate per borehole.
        Q : float
            The total heat extraction rate (in watts).
        T0 : array
            Borehole wall temperature at nodes assuming zero heat
            extraction rate (in degree Celsius).
        m_flow_small : float
            The minimum fluid mass flow rate (in kg/s). If `f_m_flow`
            returns a smaller total value, the fluid mass flow rate and
            the heat extraction rate are set to zero. If the total value
            is above `m_flow_small`, the minimum fluid mas flow rate per
            borehole is set to `m_flow_small`.

        Returns
        -------
        Q : float
            The total heat extraction rate (in watts).
        q : array
            The heat extraction rate at the nodes (in W/m).
        T_b : array
            The borehole wall temperature at the nodes (in degree
            Celsius).
        T_f_in : float
            The inlet fluid temperature (in degree Celsius).
        T_f_out : float
            The outlet fluid temperature (in degree Celsius).

        """
        m_flow = jnp.maximum(m_flow, m_flow_small)
        # Build and solve system of equations
        A, B = self.update_system_of_equations(
            m_flow,
            Q,
            T0)
        X = jnp.linalg.solve(A, B)
        # Store results
        q = X[:self.n_nodes].reshape((self.borefield.n_boreholes, -1))
        T_b = T0 - jnp.tensordot(self.h_to_self, q, axes=([-2, -1], [-2, -1]))
        T_f_in = X[-1]
        T_f_out = T_f_in + Q / (jnp.sum(m_flow) * self.cp_f)
        return Q, q, T_b, T_f_in, T_f_out

    @partial(jit, static_argnames=['self'])
    def _simulate_step_no_flow(self, m_flow: float | Array, Q: float, T0: Array, m_flow_small: float) -> Tuple[float, Array, Array, float, float]:
        """Solve a single time step with zero mass flow rate.

        Parameters
        ----------
        m_flow : float or array
            Total fluid mass flow rate (in kg/s), or array of fluid mass
            flow rate per borehole.
        Q : float
            The total heat extraction rate (in watts).
        T0 : array
            Borehole wall temperature at nodes assuming zero heat
            extraction rate (in degree Celsius).
        m_flow_small : float
            The minimum fluid mass flow rate (in kg/s). If `f_m_flow`
            returns a smaller total value, the fluid mass flow rate and
            the heat extraction rate are set to zero. If the total value
            is above `m_flow_small`, the minimum fluid mas flow rate per
            borehole is set to `m_flow_small`.

        Returns
        -------
        Q : float
            The total heat extraction rate (in watts).
        q : array
            The heat extraction rate at the nodes (in W/m).
        T_b : array
            The borehole wall temperature at the nodes (in degree
            Celsius).
        T_f_in : nan
            The inlet fluid temperature (in degree Celsius).
        T_f_out : nan
            The outlet fluid temperature (in degree Celsius).

        """
        # Store results
        Q = 0.
        q = jnp.zeros((self.borefield.n_boreholes, self.borefield.n_nodes))
        T_b = jnp.full((self.borefield.n_boreholes, self.borefield.n_nodes), T0)
        T_f_in = jnp.nan
        T_f_out = jnp.nan
        return Q, q, T_b, T_f_in, T_f_out



# -*- coding: utf-8 -*-
from typing import Tuple

from jax import numpy as jnp
from matplotlib import pyplot as plt
from matplotlib.axes import Axes
from matplotlib.figure import Figure

from ._format_axis import _format_axis
from ..simulation import gFunction


def plot_gfunction(gfunction: gFunction, ax: Axes | None = None) -> Axes | Tuple[Figure, Axes]:
    """Plot the g-function of a borefield.

    Parameters
    ----------
    gfunction : g-function
        g-Function to be plotted.
    ax : axes, None, default: ``None``
        The axis on which to draw the ground heat exchanger. If `ax` is
        ``None``, a new figure and new axes are created.

    Returns
    -------
    fig : figure
        The figure. Only return if `ax` is ``None``.
    ax : axes
        The axis.

    """
    ax_is_none = ax is None
    if ax is None:
        fig, ax = plt.subplots(layout='constrained')
    _format_axis(ax, axis_labels=[r'$\ln(t/t_s)$', r'$g$-function'])

    # Dimensionless time
    time = gfunction.time[1:]
    alpha = gfunction.alpha
    L = gfunction.borefield.L.mean()
    ts = L**2 / (9 * alpha)
    lntts = jnp.log(time / ts)

    # Evaluate g-function if not already evaluated
    try:
        gfunction.g
    except:
        gfunction.simulate()

    # Plot g-function
    ax.plot(lntts, gfunction.g)

    if ax_is_none:
        return fig, ax
    else:
        return ax



# -*- coding: utf-8 -*-
from functools import partial
from typing import Tuple

from jax import numpy as jnp
from jax import Array, jit, vmap
from jax.lax import fori_loop, switch

from ._tube import _Tube
from ..basis import Basis
from ..utilities import quadgl


class SingleUTube(_Tube):
    """Single U-Tube geothermal borehole.

    Parameters
    ----------
    R_d : array_like or callable
        (2, 2,) array of thermal resistances (in m-K/W), or callable that
        takes the mass flow rate as input (in kg/s) and returns a (2, 2,)
        array.
    r_b : float
        Borehole radius (in meters).
    path : path
        Path of the borehole.
    basis : basis
        Basis functions.
    n_segments : int
        Number of segments.
    segment_ratios : array_like or None, default: None
        Normalized size of the segments. Should total ``1``
        (i.e. ``sum(segment_ratios) = 1``). If `segment_ratios` is
        ``None``, segments of equal size are considered (i.e.
        ``segment_ratios[v] = 1 / n_segments``).
    order : int, default: 101
        Order of the Gauss-Legendre quadrature to evaluate thermal
        response factors to points outside the borehole, and to evaluate
        coefficient matrices for fluid and heat extraction rate profiles.
    order_to_self : int, default: 21
        Order of the tanh-sinh quadrature to evaluate thermal
        response factors to nodes on the borehole. Corresponds to the
        number of quadrature points along each subinterval delimited
        by nodes and edges of the segments.

    Attributes
    ----------
    n_pipes : int
        Number of pipes in the borehole.
    n_nodes : int
        Total number of nodes along the borehole.
    xi_edges : array
        (`n_segments`+1,) array of the coordinates of the edges of the
        segments.
    L : float
        Borehole length (in meters).
    xi : array
        (`n_nodes`,) array of node coordinates.
    p : array
        (`n_nodes`, 3,) array of node positions.
    dp_dxi : array
        (`n_nodes`, 3,) array of the derivatives of the position at the
        node coordinates.
    J : array
        (`n_nodes`,) array of the norm of the Jacobian at the node
        coordinates.
    s : array
        (`n_nodes`,) array of the longitudinal position at the node
        coordinates.
    w : array
        (`n_nodes`,) array of quadrature weights at the node coordinates.
        These quadrature weights take into account the norm of the
        Jacobian.

    """

    @classmethod
    @partial(jit, static_argnames=['cls'])
    def _fluid_temperature_a_in(cls, xi_p: float, index: int, beta_ij: Array, top_connectivity: Tuple[Array, Array], s_coefs: Array) -> Array:
        """Inlet coefficient to evaluate the fluid temperatures.

        Parameters
        ----------
        xi_p : float
            Coordinate along the borehole segment.
        index : int
            Index of the borehole segment.
        beta_ij : array
            (`n_pipes`, `n_pipes`,) array of thermal conductance
            coefficients.
        top_connectivity : tuple of array
            Tuple of two arrays (``c_in`` and ``c_u``) of shape
            (`n_pipes`/2,) and (`n_pipes`/2, `n_pipes`/2,). The two
            arrays give the relation between the inlet fluid temperature,
            the fluid temperatures at the top-end of the borehole in the
            upward flowing pipes, and the fluid temperatures at the top-end
            of the borehole in the downward flowing pipes following the
            relation: ``T_fd = c_in * T_f_in + c_u @ T_fu``.
        s_coefs : array
            (`n_nodes`, `n_segments`,) array of polynomial coefficients
            for the evaluation of the longitudinal position (in meters) along
            the borehole as a function of `xi_p`.

        Returns
        -------
        array
            (`n_pipes`,) array of coefficients for the inlet fluid
            temperature.

        """
        b_in = cls._outlet_fluid_temperature_a_in(
            beta_ij, top_connectivity, None, s_coefs)
        c_in = cls._general_solution_a_in(
            xi_p, index, beta_ij, s_coefs)
        c_out = cls._general_solution_a_out(
            xi_p, index, beta_ij, s_coefs)
        a_in = c_in + b_in * c_out
        return a_in

    @classmethod
    @partial(jit, static_argnames=['cls', 'order'])
    def _fluid_temperature_a_b(cls, xi_p: float, index: int, beta_ij: Array, top_connectivity: Tuple[Array, Array], s_coefs: Array, J_coefs: Array, psi_coefs: Array, order: int = 101) -> Array:
        """Borehole wall coefficient to evaluate the fluid temperatures.

        Parameters
        ----------
        xi_p : float
            Coordinate along the borehole segment.
        index : int
            Index of the borehole segment.
        beta_ij : array
            (`n_pipes`, `n_pipes`,) array of thermal conductance
            coefficients.
        top_connectivity : tuple of array
            Tuple of two arrays (``c_in`` and ``c_u``) of shape
            (`n_pipes`/2,) and (`n_pipes`/2, `n_pipes`/2,). The two
            arrays give the relation between the inlet fluid temperature,
            the fluid temperatures at the top-end of the borehole in the
            upward flowing pipes, and the fluid temperatures at the top-end
            of the borehole in the downward flowing pipes following the
            relation: ``T_fd = c_in * T_f_in + c_u @ T_fu``.
        s_coefs : array
            (`n_nodes`, `n_segments`,) array of polynomial coefficients
            for the evaluation of the longitudinal position (in meters) along
            the borehole segments as a function of `xi_p`.
        J_coefs : array
            (`n_nodes`, `n_segments`,) array of polynomial coefficients
            for the evaluation of the norm of the jacobian (in meters) along
            the borehole segments as a function of `xi_p`.
        psi_coefs : array
            (`n_nodes`, `n_nodes`,) array of polynomial coefficients
            for the evaluation of the polynomial basis functions along
            the borehole segments as a function of `xi_p`.
        order : int, default: 101
            Order of the numerical quadrature.

        Returns
        -------
        array
            (`n_pipes`, `n_nodes`,) array of coefficients for the
            borehole wall temperature.

        """
        b_b = cls._outlet_fluid_temperature_a_b(
            beta_ij, top_connectivity, None, s_coefs, J_coefs, psi_coefs, order)
        c_out = cls._general_solution_a_out(
            xi_p, index, beta_ij, s_coefs)
        c_b = cls._general_solution_a_b(
            xi_p, index, beta_ij, s_coefs, J_coefs, psi_coefs, order)
        a_b = c_b + jnp.outer(c_out, b_b)
        return a_b

    @classmethod
    @partial(jit, static_argnames=['cls'])
    def _general_solution_a_in(cls, xi_p: float, index: int, beta_ij: Array, s_coefs: Array) -> Array:
        """Inlet coefficient to evaluate the general solution.

        Parameters
        ----------
        xi_p : float
            Coordinate along the borehole segment.
        index : int
            Index of the borehole segment.
        beta_ij: array
            (2, 2,) array of thermal conductance coefficients.
        s_coefs : array
            (`n_nodes`, `n_segments`,) array of polynomial coefficients
            for the evaluation of the longitudinal position (in meters) along
            the borehole segments as a function of `xi_p`.

        Returns
        -------
        array
            (2,) array of coefficients for the inlet fluid temperature.

        """
        s = cls._longitudinal_position(xi_p, index, s_coefs)
        a_in = jnp.array(
            [cls._f1(s, beta_ij), -cls._f2(s, beta_ij)]
        )
        return a_in

    @classmethod
    @partial(jit, static_argnames=['cls'])
    def _general_solution_a_out(cls, xi_p: float, index: int, beta_ij: Array, s_coefs: Array) -> Array:
        """Outlet coefficient to evaluate the general solution.

        Parameters
        ----------
        xi_p : float
            Coordinate along the borehole segment.
        index : int
            Index of the borehole segment.
        beta_ij: array
            (2, 2,) array of thermal conductance coefficients.
        s_coefs : array
            (`n_nodes`, `n_segments`,) array of polynomial coefficients
            for the evaluation of the longitudinal position (in meters) along
            the borehole segments as a function of `xi_p`.

        Returns
        -------
        array
            (2,) array of coefficients for the outlet fluid
            temperature.

        """
        s = cls._longitudinal_position(xi_p, index, s_coefs)
        a_out = jnp.array(
            [cls._f2(s, beta_ij), cls._f3(s, beta_ij)]
        )
        return a_out

    @classmethod
    @partial(jit, static_argnames=['cls', 'order'])
    def _general_solution_a_b(cls, xi_p: float, index: int, beta_ij: Array, s_coefs: Array, J_coefs: Array, psi_coefs: Array, order: int = 101) -> Array:
        """Borehole wall coefficient to evaluate the general solution.

        Parameters
        ----------
        xi_p : float
            Coordinate along the borehole segment.
        index : int
            Index of the borehole segment.
        beta_ij: array
            (2, 2,) array of thermal conductance coefficients.
        s_coefs : array
            (`n_nodes`, `n_segments`,) array of polynomial coefficients
            for the evaluation of the longitudinal position (in meters) along
            the borehole segments as a function of `xi_p`.
        J_coefs : array
            (`n_nodes`, `n_segments`,) array of polynomial coefficients
            for the evaluation of the norm of the jacobian (in meters) along
            the borehole segments as a function of `xi_p`.
        psi_coefs : array
            (`n_nodes`, `n_nodes`,) array of polynomial coefficients
            for the evaluation of the polynomial basis functions along
            the borehole segments as a function of `xi_p`.
        order : int, default: 101
            Order of the numerical quadrature.

        Returns
        -------
        array
            (2, `n_nodes`,) array of coefficients for the borehole wall
            temperature.

        """
        # Initialize coefficients
        n_pipes = jnp.shape(beta_ij)[0]
        n_segments = jnp.shape(s_coefs)[1]
        n_nodes = jnp.shape(psi_coefs)[0]
        a_b = jnp.zeros((n_pipes, n_segments, n_nodes))
        # Longitudinal position corresponding to coordinate `xi_p`
        s = cls._longitudinal_position(xi_p, index, s_coefs)

        def general_solution_integrand(_xi_p: float, _i: int) -> Array:
            """Integrand of the general solution."""
            # Longitudinal positions, norms of Jacobian and basis functions
            # at integration points
            _t = cls._longitudinal_position(_xi_p, _i, s_coefs)
            _J = cls._norm_of_jacobian(_xi_p, _i, J_coefs)
            _psi = Basis._f_psi(_xi_p, psi_coefs)
            # Integrand
            _integrand = _J * jnp.stack(
                [
                    _psi * cls._f4(s - _t, beta_ij),
                    -_psi * cls._f5(s - _t, beta_ij)
                ]
            )
            return _integrand

        def _integral(_i: int, _b: float) -> Array:
            """Integrals over a portion of a segment."""
            _a_b = quadgl(
                partial(general_solution_integrand, _i=_i),
                points=jnp.array([-1., _b]),
                order=order
            )
            return _a_b

        def _zeros(_i: int) -> Array:
            """Array of zeros for segments below `index`."""
            return jnp.zeros((n_pipes, n_nodes))

        branches = [
            partial(_integral, _b=1.),
            partial(_integral, _b=xi_p),
            _zeros
        ]

        # Evaluation of the integral along all segments
        def _evaluate_integrals(_i, _a_b):
            _a_b = _a_b.at[:, _i, :].set(switch(_i - index + 1, branches, _i))
            return _a_b
        a_b = fori_loop(0, n_segments, _evaluate_integrals, a_b, unroll=False)

        return a_b.reshape(n_pipes, -1)

    @classmethod
    @partial(jit, static_argnames=['cls'])
    def _outlet_fluid_temperature_a_in(cls, beta_ij: Array, top_connectivity: Tuple[Array, Array], mixing: Array, s_coefs: Array) -> float:
        """Inlet coefficient to evaluate the outlet fluid temperature.

        Parameters
        ----------
        beta_ij: array
            (2, 2,) array of thermal conductance coefficients.
        top_connectivity : tuple of array
            Tuple of two arrays (``c_in`` and ``c_u``) of shape
            (`n_pipes`/2,) and (`n_pipes`/2, `n_pipes`/2,). The two
            arrays give the relation between the inlet fluid temperature,
            the fluid temperatures at the top-end of the borehole in the
            upward flowing pipes, and the fluid temperatures at the top-end
            of the borehole in the downward flowing pipes following the
            relation: ``T_fd = c_in * T_f_in + c_u @ T_fu``.
        mixing : array
            (`n_pipes`/2,) array of coefficients to evaluate the outlet fluid
            temperature from the fluid temperatures at the top-end of the
            borehole in the upward flowing pipes following the relation:
            ``T_f_out = mixing @ T_fu``.
        s_coefs : array
            (`n_nodes`, `n_segments`,) array of polynomial coefficients
            for the evaluation of the longitudinal position (in meters) along
            the borehole segments as a function of `xi_p`.

        Returns
        -------
        float
            Coefficient for the inlet fluid temperature.

        """
        n_segments = jnp.shape(s_coefs)[1]
        L = cls._longitudinal_position(1., n_segments, s_coefs)
        a_in = (
            cls._f1(L, beta_ij) + cls._f2(L, beta_ij)
        ) / (cls._f3(L, beta_ij) - cls._f2(L, beta_ij))
        return a_in

    @classmethod
    @partial(jit, static_argnames=['cls', 'order'])
    def _outlet_fluid_temperature_a_b(cls, beta_ij: Array, top_connectivity: Tuple[Array, Array], mixing: Array, s_coefs: Array, J_coefs: Array, psi_coefs: Array, order: int = 101) -> Array:
        """Borehole coefficient to evaluate the outlet fluid temperature.

        Parameters
        ----------
        beta_ij: array
            (2, 2,) array of thermal conductance coefficients.
        top_connectivity : tuple of array
            Tuple of two arrays (``c_in`` and ``c_u``) of shape
            (`n_pipes`/2,) and (`n_pipes`/2, `n_pipes`/2,). The two
            arrays give the relation between the inlet fluid temperature,
            the fluid temperatures at the top-end of the borehole in the
            upward flowing pipes, and the fluid temperatures at the top-end
            of the borehole in the downward flowing pipes following the
            relation: ``T_fd = c_in * T_f_in + c_u @ T_fu``.
        mixing : array
            (`n_pipes`/2,) array of coefficients to evaluate the outlet fluid
            temperature from the fluid temperatures at the top-end of the
            borehole in the upward flowing pipes following the relation:
            ``T_f_out = mixing @ T_fu``.
        s_coefs : array
            (`n_nodes`, `n_segments`,) array of polynomial coefficients
            for the evaluation of the longitudinal position (in meters) along
            the borehole segments as a function of `xi_p`.
        J_coefs : array
            (`n_nodes`, `n_segments`,) array of polynomial coefficients
            for the evaluation of the norm of the jacobian (in meters) along
            the borehole segments as a function of `xi_p`.
        psi_coefs : array
            (`n_nodes`, `n_nodes`,) array of polynomial coefficients
            for the evaluation of the polynomial basis functions along
            the borehole segments as a function of `xi_p`.
        order : int, default: 101
            Order of the numerical quadrature.

        Returns
        -------
        array
            (`n_nodes`,) array of coefficients for the borehole wall
            temperature.

        """
        n_segments = jnp.shape(s_coefs)[1]
        L = cls._longitudinal_position(1., n_segments, s_coefs)
        one_over_f3_minus_f2 = 1 / (cls._f3(L, beta_ij) - cls._f2(L, beta_ij))

        def outlet_fluid_temperature_integrand(_xi_p: float, _i: int) -> Array:
            """Integrand function."""
            # Longitudinal positions, norms of Jacobian and basis functions
            # at integration points
            _t = cls._longitudinal_position(_xi_p, _i, s_coefs)
            _J = cls._norm_of_jacobian(_xi_p, _i, J_coefs)
            _psi = Basis._f_psi(_xi_p, psi_coefs)
            # Integrand
            _integrand = _psi * _J * one_over_f3_minus_f2 * (
                cls._f4(L - _t, beta_ij)
                + cls._f5(L - _t, beta_ij)
            )
            return _integrand

        def outlet_fluid_temperature_integral(_i: int) -> Array:
            """Integrals over a segment."""
            _integral = quadgl(
                partial(outlet_fluid_temperature_integrand, _i= _i),
                order=order
            )
            return _integral

        a_b = vmap(
            outlet_fluid_temperature_integral,
            in_axes=0,
            out_axes=0
        )(jnp.arange(n_segments))

        return a_b.flatten()

    @staticmethod
    def _beta(beta_ij: Array) -> float:
        """Coefficient ``beta`` from Hellström (1991).

        Parameters
        ----------
        beta_ij: array
            (2, 2,) array of thermal conductance coefficients.

        Returns
        -------
        float

        """
        beta1 = beta_ij[0, 0]
        beta2 = beta_ij[1, 1]
        beta = 0.5 * (beta2 - beta1)
        return beta

    @staticmethod
    def _gamma(beta_ij: Array) -> float:
        """Coefficient ``gamma`` from Hellström (1991).

        Parameters
        ----------
        beta_ij: array
            (2, 2,) array of thermal conductance coefficients.

        Returns
        -------
        float

        """
        beta1 = beta_ij[0, 0]
        beta2 = beta_ij[1, 1]
        beta12 = beta_ij[0, 1]
        gamma = jnp.sqrt(0.25 * (beta1 + beta2)**2 + beta12 * (beta1 + beta2))
        return gamma

    @classmethod
    @partial(jit, static_argnames=['cls'])
    def _delta(cls, beta_ij: Array) -> float:
        """Coefficient ``delta`` from Hellström (1991).

        Parameters
        ----------
        beta_ij: array
            (2, 2,) array of thermal conductance coefficients.

        Returns
        -------
        float

        """
        beta1 = beta_ij[0, 0]
        beta2 = beta_ij[1, 1]
        beta12 = beta_ij[0, 1]
        gamma = cls._gamma(beta_ij)
        delta = 1. / gamma * (beta12 + 0.5 * (beta1 + beta2))
        return delta

    @classmethod
    @partial(jit, static_argnames=['cls'])
    def _f1(cls, s: Array | float, beta_ij: Array) -> Array | float:
        """Function ``f1`` from Hellström (1991).

        Parameters
        ----------
        s : array or float
            (M,) array of longitudinal positions (in meters).
        beta_ij: array
            (2, 2,) array of thermal conductance coefficients.

        Returns
        -------
        array or float
            (M,) array.

        """
        beta = cls._beta(beta_ij)
        gamma = cls._gamma(beta_ij)
        delta = cls._delta(beta_ij)
        f1 = jnp.exp(beta * s) * (jnp.cosh(gamma * s) - delta * jnp.sinh(gamma * s))
        return f1

    @classmethod
    @partial(jit, static_argnames=['cls'])
    def _f2(cls, s: Array | float, beta_ij: Array) -> Array | float:
        """Function ``f2`` from Hellström (1991).

        Parameters
        ----------
        s : array or float
            (M,) array of longitudinal positions (in meters).
        beta_ij: array
            (2, 2,) array of thermal conductance coefficients.

        Returns
        -------
        array or float
            (M,) array.

        """
        beta = cls._beta(beta_ij)
        gamma = cls._gamma(beta_ij)
        beta12 = beta_ij[0, 1]
        f2 = jnp.exp(beta * s) * beta12 / gamma * jnp.sinh(gamma * s)
        return f2

    @classmethod
    @partial(jit, static_argnames=['cls'])
    def _f3(cls, s: Array | float, beta_ij: Array) -> Array | float:
        """Function ``f3`` from Hellström (1991).

        Parameters
        ----------
        s : array or float
            (M,) array of longitudinal positions (in meters).
        beta_ij: array
            (2, 2,) array of thermal conductance coefficients.

        Returns
        -------
        array or float
            (M,) array.

        """
        beta = cls._beta(beta_ij)
        gamma = cls._gamma(beta_ij)
        delta = cls._delta(beta_ij)
        f3 = jnp.exp(beta * s) * (jnp.cosh(gamma * s) + delta * jnp.sinh(gamma * s))
        return f3

    @classmethod
    @partial(jit, static_argnames=['cls'])
    def _f4(cls, s: Array | float, beta_ij: Array) -> Array | float:
        """Function ``f4`` from Hellström (1991).

        Parameters
        ----------
        s : array or float
            (M,) array of longitudinal positions (in meters).
        beta_ij: array
            (2, 2,) array of thermal conductance coefficients.

        Returns
        -------
        array or float
            (M,) array.

        """
        beta1 = beta_ij[0, 0]
        beta2 = beta_ij[1, 1]
        beta12 = beta_ij[0, 1]
        beta = cls._beta(beta_ij)
        gamma = cls._gamma(beta_ij)
        delta = cls._delta(beta_ij)
        f4 = jnp.exp(beta * s) * (beta1 * jnp.cosh(gamma * s) - (delta * beta1 + beta2 * beta12 / gamma) * jnp.sinh(gamma * s))
        return f4

    @classmethod
    @partial(jit, static_argnames=['cls'])
    def _f5(cls, s: Array | float, beta_ij: Array) -> Array | float:
        """Function ``f5`` from Hellström (1991).

        Parameters
        ----------
        s : array or float
            (M,) array of longitudinal positions (in meters).
        beta_ij: array
            (2, 2,) array of thermal conductance coefficients.

        Returns
        -------
        array or float
            (M,) array.

        """
        beta1 = beta_ij[0, 0]
        beta2 = beta_ij[1, 1]
        beta12 = beta_ij[0, 1]
        beta = cls._beta(beta_ij)
        gamma = cls._gamma(beta_ij)
        delta = cls._delta(beta_ij)
        f5 = jnp.exp(beta * s) * (beta2 * jnp.cosh(gamma * s) + (delta * beta2 + beta1 * beta12 / gamma) * jnp.sinh(gamma * s))
        return f5



from .conduction_thermal_resistance import conduction_thermal_resistance_circular_pipe
from .convective_heat_transfer_coefficient import (
convective_heat_transfer_coefficient_annular_pipe,
convective_heat_transfer_coefficient_circular_pipe,
nusselt_number_annular_pipe,
nusselt_number_circular_pipe
)
from .darcy_friction_factor import darcy_friction_factor
from .reynolds_number import (
reynolds_number_annular_pipe,
reynolds_number_circular_pipe
)
from .multipole import Multipole



# -*- coding: utf-8 -*-
from functools import partial
from time import perf_counter

from jax import numpy as jnp
from jax import Array, jit, vmap
from jax.typing import ArrayLike
import numpy as np

from ..borefield.borefield import Borefield
from ..borefield.network import Network
from ._temporal_superposition import _TemporalSuperposition


class LoadAggregation(_TemporalSuperposition):
    """Load aggregation.

    Parameters
    ----------
    borefield : borefield or network
        The borefield.
    dt : float
        Time step (in seconds).
    tmax : float
        Maximum time of the simulation (in seconds).
    alpha : float
        Ground thermal diffusivity (in m^2/s).
    cells_per_level : int, default: ``5``
        Number of cells per aggregation level.
    p : array_like or None, default: ``None``
        (`n_points`, 3,) array of positions to evaluate the ground
        temperature. If `p` is ``None``, the ground temperature is not
        evaluated.
    disp : bool, default: ``True``
        Set to ``True`` to print initialization progression messages.

    Attributes
    ----------
    n_cells : int
        Number of aggregation cells.
    n_points : int
        Number of points to evaluate ground temperatures.
    time : array
        (`n_cells`+1,) array of start/end times of the aggregation cells
        (in seconds).
    A : array
        Load shifting matrix.
    h_to_self : array
        (`n_cells`, `n_boreholes`, `n_nodes`, `n_boreholes`,
        `n_nodes`,) array of thermal response factors at the nodes.
    h_to_point : array
        (`n_cells`, `n_points`, `n_boreholes`, `n_nodes`,) array of
        thermal response factors at points.

    """

    def __init__(self, borefield: Borefield | Network, dt: float, tmax: float, alpha: float, cells_per_level: int = 5, p: ArrayLike | None = None, disp: bool = True):
        tic = perf_counter()
        # Runtime type validation
        if not isinstance(p, ArrayLike) and p is not None:
            raise TypeError(f"Expected arraylike or None input; got {p}")
        # Convert input to jax.Array
        if p is not None:
            p = jnp.asarray(p)

        self.borefield = borefield
        self.dt = dt
        self.tmax = tmax
        self.cells_per_level = cells_per_level
        self.p = p
        self._time = 0.
        self._k = -1
        self.time = self._load_aggregation_cells(dt, tmax, cells_per_level)
        self.n_cells = len(self.time) - 1
        self.A = self._load_shifting_matrix(self.time)
        self.q = jnp.zeros((len(self.time) - 1, borefield.n_boreholes, borefield.n_nodes))
        if disp:
            print('Initialization start.')
        self.h_to_self = borefield.h_to_self(self.time[1:], alpha)
        self.h_to_self = self.h_to_self.at[1:].set(jnp.diff(self.h_to_self, axis=0))
        if disp:
            toc = perf_counter()
            print(
                f'Completed thermal response factors to nodes. '
                f'Elapsed time: {toc-tic:.2f} seconds.'
                    )
        if p is not None:
            self.n_points = p.shape[0]
            self.h_to_point = borefield.h_to_point(p, self.time[1:], alpha)
            self.h_to_point = self.h_to_point.at[1:].set(jnp.diff(self.h_to_point, axis=0))
            if disp:
                toc = perf_counter()
                print(
                    f'Completed thermal response factors to ground. '
                    f'Elapsed time: {toc-tic:.2f} seconds.'
                        )
        else:
            self.n_points = 0
            self.h_to_point = jnp.zeros((0, borefield.n_boreholes, borefield.n_nodes))
        if disp:
            toc = perf_counter()
            print(
                f'Initialization end. Elapsed time: {toc-tic:.2f} seconds.'
            )

    def next_time_step(self) -> float:
        """Advance to next simulation time step.

        Returns
        -------
        float
            Time (in seconds) of the new time step.

        """
        self.q = self._next_time_step(self.A, self.q)
        self._time += self.dt
        return self._time

    def reset_history(self):
        """Reset the history to its initial condition.

        """
        self.q = self.q.at[:].set(0.)
        self._time = 0.
        self._k = -1

    def set_current_load(self, q: Array):
        """Set the current heat extraction rate.

        Parameters
        ----------
        q : array
            (`n_boreholes`, `n_nodes`,) array of heat extraction rates
            (in W/m).

        """
        self.q = self._current_load(self.q, q)

    def temperature(self) -> Array:
        """Evaluate temperatures at nodes.

        Returns
        ----------
        T_b : array
            (`n_boreholes`, `n_nodes`,) array of borehole wall
            temperature variations (in degree Celcius).

        """
        T_b = self._temperature(self.h_to_self, self.q)
        return T_b

    def temperature_to_point(self) -> Array:
        """Evaluate temperatures at point.

        Returns
        ----------
        T : array
            (`n_points`,) array of ground temperature variations
            (in degree Celcius).

        """
        T = self._temperature_to_point(self.h_to_point, self.q)
        return T

    @staticmethod
    @jit
    def _current_load(q_history: Array, q: Array) -> Array:
        """Add current load to load history

        Parameters
        ----------
        q_history : array
            (`n_cells`, `n_boreholes`, `n_nodes`,) array of aggregated
            loads (in W/m).
        q : array
            (`n_boreholes`, `n_nodes`,) array of current loads (in W/m).

        Returns
        -------
        array
            (`n_cells`, `n_boreholes`, `n_nodes`,) array of updated
            aggregated loads (in W/m).

        """
        return q_history.at[0].set(q)

    @staticmethod
    @jit
    def _next_time_step(A: Array, q: Array) -> Array:
        """Shift load history.

        Parameters
        ----------
        A : array
            Load shifting matrix.
        q : array
            (`n_cells`, `n_boreholes`, `n_nodes`,) array of aggregated
            loads in (W/m).

        Returns
        -------
        array
            (`n_cells`, `n_boreholes`, `n_nodes`,) array of aggregated
            loads at next time step (in W/m).

        """
        return jnp.tensordot(A, q, axes=(1, 0))

    @staticmethod
    def _load_aggregation_cells(dt: float, tmax: float, cells_per_level: int) -> Array:
        """Build load aggregation cells.

        Parameters
        ----------
        dt : float
            Time step (in seconds).
        tmax : float
            Maximum time of the simulation (in seconds).
        cells_per_level : int, default: ``5``
            Number of cells per aggregation level.

        Returns
        -------
        array
            Array of start/end times of the aggregation cells
            (in seconds).

        """
        n_time_steps = np.ceil(tmax / dt).astype(int)
        n_levels = np.ceil(np.log2(1 + np.ceil(n_time_steps / cells_per_level))).astype(int)
        n_cells = n_levels * cells_per_level
        time = jnp.zeros(n_cells + 1)
        for i in range(1, n_cells + 1):
            # Cell size doubles every (cells_per_level) time steps
            v = jnp.ceil(i / cells_per_level)
            width = 2**(v - 1)
            time = time.at[i].set(time[i - 1] + width * dt)
        return time

    @staticmethod
    def _load_shifting_matrix(time: Array) -> Array:
        """Load shifting matrix

        Parameters
        ----------
        time : array
            Array of start/end times of the aggregation cells
            (in seconds).

        Returns
        -------
        array
            Load shifting matrix.
        """
        width = jnp.diff(time) / time[1]
        n_times = len(width)
        A = (1. - 1. / width) * jnp.eye(n_times) + jnp.diag(1. / width[1:], k=-1)
        return A



# -*- coding: utf-8 -*-
from functools import partial
from typing import Self, Tuple

from jax import numpy as jnp
from jax import Array, jit, vmap
from jax.typing import ArrayLike

from ..basis import Basis
from ..heat_transfer import point_heat_source
from ..path import Path
from ..utilities import quad


class Borehole:
    """Geothermal borehole.

    Parameters
    ----------
    r_b : float
        Borehole radius (in meters).
    path : path
        Path of the borehole.
    basis : basis
        Basis functions.
    n_segments : int
        Number of segments.
    segment_ratios : array_like or None, default: None
        Normalized size of the segments. Should total ``1``
        (i.e. ``sum(segment_ratios) = 1``). If `segment_ratios` is
        ``None``, segments of equal size are considered (i.e.
        ``segment_ratios[v] = 1 / n_segments``).
    deg : int or None, default: None
        Polynomial degree at which to approximate the trajectory of
        the borehole. If None, `deg` is either set to `path.deg` (if
        it is not None) or to `n_nodes`. The longitudinal position along
        the borehole is approximated using a degree ``2 * `deg` - 1``.
    order : int, default: 101
        Order of the Gauss-Legendre quadrature to evaluate thermal
        response factors to points outside the borehole, and to evaluate
        coefficient matrices for fluid and heat extraction rate profiles.
    order_to_self : int, default: 21
        Order of the tanh-sinh quadrature to evaluate thermal
        response factors to nodes on the borehole. Corresponds to the
        number of quadrature points along each subinterval delimited
        by nodes and edges of the segments.

    Attributes
    ----------
    n_nodes : int
        Total number of nodes along the borehole.
    xi_edges : array
        (`n_segments`+1,) array of the coordinates of the edges of the
        segments.
    L : float
        Borehole length (in meters).
    xi : array
        (`n_nodes`,) array of node coordinates.
    p : array
        (`n_nodes`, 3,) array of node positions.
    dp_dxi : array
        (`n_nodes`, 3,) array of the derivatives of the position at the
        node coordinates.
    J : array
        (`n_nodes`,) array of the norm of the Jacobian at the node
        coordinates.
    s : array
        (`n_nodes`,) array of the longitudinal position at the node
        coordinates.
    w : array
        (`n_nodes`,) array of quadrature weights at the node coordinates.
        These quadrature weights take into account the norm of the
        Jacobian.

    """

    def __init__(self, r_b: float, path: Path, basis: Basis, n_segments: int, segment_ratios: ArrayLike | None = None, deg: int | None = None, order: int = 101, order_to_self: int = 21):
        # Runtime type validation
        if not isinstance(segment_ratios, ArrayLike) and segment_ratios is not None:
            raise TypeError(f"Expected arraylike or None input; got {segment_ratios}")
        # Convert input to jax.Array
        if segment_ratios is None:
            segment_ratios = jnp.full(n_segments, 1. / n_segments)
        else:
            segment_ratios = jnp.asarray(segment_ratios)

        # --- Class attributes ---
        self.r_b = r_b
        self.path = path
        self.basis = basis
        self.n_segments = n_segments
        self.n_nodes = basis.n_nodes * n_segments
        self.segment_ratios = segment_ratios
        if deg is None:
            if path.deg is None:
                deg = basis.n_nodes
            else:
                deg = path.deg
        self.deg = deg
        self.order = order
        self.order_to_self = order_to_self
        # Segment edges
        xi_edges = 2. * jnp.cumulative_sum(segment_ratios, include_initial=True) - 1.
        self.xi_edges = xi_edges
        # Borehole length
        self.L = jnp.diff(path.f_s(jnp.array([-1., 1.])))[0]

        # --- Path functions ---
        self._initialize_path_coefficients(deg)

        # --- Nodal values of path and basis functions ---
        # Borehole coordinates (xi)
        xi = vmap(
            self.f_xi_sb,
            in_axes=(None, 0),
            out_axes=0
        )(basis.xi, jnp.arange(self.n_segments)).flatten()
        self.xi = xi
        # Positions (p)
        self.p = self.f_p(xi)
        # Derivatives of position (dp/dxi)
        self.dp_dxi = self.f_dp_dxi(xi)
        # Norms of the Jacobian (J)
        self.J = self.f_J(xi)
        # Longitudinal positions (s)
        self.s = self.f_s(xi)
        # Integration weights
        self.w = (jnp.tile(basis.w, (n_segments, 1)).T * segment_ratios).T.flatten() * self.J

    def f(self, xi: float | Array, f_nodes: Array) -> float | Array:
        """Value at coordinate from values at nodes.

        Parameters
        ----------
        xi : float or array
            Borehole coordinate or (M,) array of borehole coordinates.
        f_nodes : array
            (`n_nodes`,) array of values at borehole nodes.

        Returns
        -------
        float or array
            Value or (M,) array of values at requested coordinates
            evaluated using polynomial basis functions.

        """
        if len(jnp.shape(xi)) > 0:
            return vmap(self.f, in_axes=(0, None), out_axes=0)(xi, f_nodes)
        xi_p, index = self.f_xi_bs(xi)
        f_nodes_segment = f_nodes.reshape(self.n_segments, -1)[index]
        return self.basis.f(xi_p, f_nodes_segment)

    def f_dp_dxi(self, xi: float | Array) -> Array:
        """Derivative of the position along the borehole.

        Parameters
        ----------
        xi : float or array
            Coordinate or (M,) array of coordinates along the borehole.

        Returns
        -------
        float
            (M,) or (M, 3,) array of derivatives of the position along the
            borehole (in meters).

        """
        if len(jnp.shape(xi)) > 0:
            return vmap(
                self.f_dp_dxi,
                in_axes=0,
                out_axes=0
            )(xi)
        xi_p, index = self.f_xi_bs(xi)
        return self._derivative_of_position(xi_p, index, self._dp_dxi_coefs) / self.segment_ratios[index]

    def f_J(self, xi: float | Array) -> float | Array:
        """Norm of the Jacobian along the borehole.

        Parameters
        ----------
        xi : float or array
            Coordinate or (M,) array of coordinates along the borehole.

        Returns
        -------
        float
            Norm of the Jacobian or (M,) array of norms of the Jacobian
            (in meters).

        """
        if len(jnp.shape(xi)) > 0:
            return vmap(
                self.f_J,
                in_axes=0,
                out_axes=0
            )(xi)
        xi_p, index = self.f_xi_bs(xi)
        return self._norm_of_jacobian(xi_p, index, self._J_coefs) / self.segment_ratios[index]

    def f_p(self, xi: float | Array) -> Array:
        """Position along the borehole.

        Parameters
        ----------
        xi : float or array
            Coordinate or (M,) array of coordinates along the borehole.

        Returns
        -------
        float
            (3,) or (M, 3,) array of positions along the borehole
            (in meters).

        """
        if len(jnp.shape(xi)) > 0:
            return vmap(
                self.f_p,
                in_axes=0,
                out_axes=0
            )(xi)
        xi_p, index = self.f_xi_bs(xi)
        return self._position(xi_p, index, self._p_coefs)

    def f_s(self, xi: float | Array) -> float | Array:
        """Longitudinal position along the borehole.

        Parameters
        ----------
        xi : float or array
            Coordinate or (M,) array of coordinates along the borehole.

        Returns
        -------
        float
            Longitudinal position or (M,) array of longitudinal
            positions along the borehole (in meters).

        """
        if len(jnp.shape(xi)) > 0:
            return vmap(
                self.f_s,
                in_axes=0,
                out_axes=0
            )(xi)
        xi_p, index = self.f_xi_bs(xi)
        return self._longitudinal_position(xi_p, index, self._s_coefs)

    def f_xi_bs(self, xi: float | Array) -> Tuple[float | Array, int]:
        """Segment coordinate from borehole coordinate.

        Parameters
        ----------
        xi : float or array
            Coordinate or (M,) array of coordinates along the borehole.

        Returns
        -------
        xi_p : float or array
            Coordinate or (M,) array of coordinates along the borehole
            segments.
        index : int or array
            Index or (M,) array of indices of the borehole segments.

        """
        if len(jnp.shape(xi)) > 0:
            return vmap(
                self.f_xi_bs,
                in_axes=0,
                out_axes=0
            )(xi)
        return self._segment_coordinate(xi, self.xi_edges)

    def f_xi_sb(self, xi_p: float, index: int) -> float:
        """Borehole coordinate from segment coordinate.

        Parameters
        ----------
        xi_p : float
            Coordinate along the borehole segment.
        index : int
            Index of the borehole segment.

        Returns
        -------
        xi : float
            Coordinate along the borehole.

        """
        return self._borehole_coordinate(xi_p, self.xi_edges, index)

    def h_to_borehole(self, borehole: Self, time: ArrayLike, alpha: float) -> Array:
        """Thermal response factors to nodes of another borehole.

        Parameters
        ----------
        borehole : borehole
            Borehole for which the thermal response factors will be
            evaluated at the nodes.
        time : array
            (K,) array of times (in seconds).
        alpha : float
            Ground thermal diffusivity (in m^2/s).

        Returns
        -------
        array
            (K, ```borehole`.n_nodes``, `n_nodes`,) array of thermal
            response factors.
        """
        return self.h_to_point(borehole.p, time, alpha)

    def h_to_coordinate_on_self(self, xi: Array, time: Array, alpha: float) -> Array:
        """Thermal response factors to coordinates along itself.

        Parameters
        ----------
        xi : array
            (M,) array of the coordinates along itself at which thermal
            response factors are evaluated.
        time : array
            (K,) array of times (in seconds).
        alpha : float
            Ground thermal diffusivity (in m^2/s).

        Returns
        -------
        array
            (K, M, `n_nodes`,) array of thermal response factors.
        """
        # Positions (p) of points on self
        p = self.path.f_p(xi)
        return self.h_to_point(p, time, alpha, r_min=self.r_b)

    def h_to_point(self, p: Array, time: Array, alpha: float, r_min: float = 0.):
        """Thermal response factors to a point.

        Parameters
        ----------
        p : array
            (M, 3,) array of the positions at which thermal response
            factors are evaluated.
        time : array
            (K,) array of times (in seconds).
        alpha : float
            Ground thermal diffusivity (in m^2/s).
        r_min : float, default: ``0.``
            Minimum distance (in meters) added to the distance between the
            heat source and the positions `p`.

        Returns
        -------
        array
            (K, M, `n_nodes`,) array of thermal response factors.
        """
        n_nodes = self.n_nodes
        if len(jnp.shape(time)) > 0:
            n_times = len(time)
            if len(jnp.shape(p)) > 1:
                shape = (n_times, -1, n_nodes)
            else:
                shape = (n_times, n_nodes)
            h_to_point = vmap(
                self.h_to_point,
                in_axes=(None, 0, None, None)
            )(
                p, time, alpha, r_min
            ).reshape(shape)
            return h_to_point
        if len(jnp.shape(p)) > 1:
            h_to_point = vmap(
                self.h_to_point,
                in_axes=(0, None, None, None)
            )(
                p, time, alpha, r_min
            ).reshape(-1, n_nodes)
            return h_to_point
        h_to_point = self._h_to_point(p, time, alpha, r_min)
        return h_to_point

    def h_to_self(self, time: Array, alpha: float):
        """Thermal response factors to its own nodes.

        Parameters
        ----------
        time : array
            (K,) array of times (in seconds).
        alpha : float
            Ground thermal diffusivity (in m^2/s).

        Returns
        -------
        array
            (K, `n_nodes`, `n_nodes`,) array of thermal response factors.
        """
        n_nodes = self.n_nodes
        if len(jnp.shape(time)) > 0:
            n_times = len(time)
            h_to_self = vmap(
                self.h_to_self,
                in_axes=(0, None)
            )(
                time, alpha
            )
            return h_to_self
        h_to_self = vmap(
            self._h_to_node_on_self,
            in_axes=(0, None, None)
        )(
            self.p, time, alpha
        ).reshape(n_nodes, n_nodes)
        return h_to_self

    def _h_to_node_on_self(self, p: Array, time: float, alpha: float):
        """Thermal response factors to its own nodes.

        Parameters
        ----------
        p : array
            (3,) array of the position of a node on the borehole at
            which thermal response factors are evaluated.
        time : float
            Time (in seconds).
        alpha : float
            Ground thermal diffusivity (in m^2/s).

        Returns
        -------
        array
            (`n_nodes`,) array of thermal response factors.
        """
        h_to_node = vmap(
            self._thermal_response_factor_to_self,
            in_axes=(0, None, None, None, None, None, None, None, None, None),
            out_axes=0
        )(
            jnp.arange(self.n_segments),
            p,
            time,
            alpha,
            self.r_b,
            self._p_coefs,
            self._J_coefs,
            self.basis._psi_coefs,
            self.basis.xi,
            self.order_to_self
        )
        return h_to_node.flatten()

    def _h_to_point(self, p: Array, time: float, alpha: float, r_min: float = 0.):
        """Thermal response factors to a point.

        Parameters
        ----------
        p : array
            (3,) array of the position at which thermal response factors
            are evaluated.
        time : float
            Time (in seconds).
        alpha : float
            Ground thermal diffusivity (in m^2/s).
        r_min : float, default: ``0.``
            Minimum distance (in meters) added to the distance between the
            heat source and the positions `p`.

        Returns
        -------
        array
            (`n_nodes`,) array of thermal response factors.
        """
        h_to_node = vmap(
            self._thermal_response_factor,
            in_axes=(0, None, None, None, None, None, None, None, None),
            out_axes=0
        )(
            jnp.arange(self.n_segments),
            p,
            time,
            alpha,
            r_min,
            self._p_coefs,
            self._J_coefs,
            self.basis._psi_coefs,
            self.order
        )
        return h_to_node.flatten()

    def _initialize_path_coefficients(self, deg: int):
        """Initialize path coefficients.

        Parameters
        ----------
        deg : int
            Polynomial degree to approximate the path along each
            segment. The longitudinal position ``s`` will be approximated
            using degree ``2 * deg - 1`` polynomials.

        """
        xi_p = jnp.linspace(-1, 1, num=deg+1)
        xi = vmap(
            self.f_xi_sb, 
            in_axes=(None, 0),
            out_axes=1
        )(xi_p, jnp.arange(self.n_segments))
        p = vmap(
            self.path.f_p,
            in_axes=1,
            out_axes=1
        )(xi)
        self._p_coefs = vmap(
            vmap(
                jnp.polyfit,
                in_axes=(None, 1, None),
                out_axes=1),
            in_axes=(None, 2, None),
            out_axes=2
        )(xi_p, p, deg)
        self._dp_dxi_coefs = vmap(
            vmap(
                jnp.polyder,
                in_axes=1,
                out_axes=1),
            in_axes=2,
            out_axes=2
        )(self._p_coefs)
        xi_p = jnp.linspace(-1, 1, num=2*deg)
        xi = vmap(
            self.f_xi_sb, 
            in_axes=(None, 0),
            out_axes=1
        )(xi_p, jnp.arange(self.n_segments))
        s = vmap(
            self.path.f_s,
            in_axes=1,
            out_axes=1
        )(xi)
        self._s_coefs = vmap(
            jnp.polyfit,
            in_axes=(None, 1, None),
            out_axes=1
        )(xi_p, s, 2*deg-1)
        self._J_coefs = vmap(
            jnp.polyder,
            in_axes=1,
            out_axes=1
        )(self._s_coefs)

    @staticmethod
    @jit
    def _borehole_coordinate(xi_p: float, xi_edges: Array, index: int) -> float:
        """Borehole coordinate from segment coordinate.

        Parameters
        ----------
        xi_p : float
            Coordinate along the borehole segment.
        xi_edges : array
            (`n_segments`+1,) array of the coordinates of the edges of
            the segments.
        index : int
            Index of the borehole segment.

        Returns
        -------
        xi : float
            Coordinate along the borehole.

        """
        a, b = xi_edges[index], xi_edges[index + 1]
        xi = 0.5 * (b + a) + 0.5 * xi_p * (b - a)
        return xi

    @staticmethod
    @jit
    def _derivative_of_position(xi_p: float, index: int, dp_dxi_coefs: Array) -> float:
        """Derivative of the position along a borehole segment.

        Parameters
        ----------
        xi_p : float
            Coordinate along the borehole segment.
        index : int
            Index of the borehole segment.
        dp_dxi_coefs : array
            (`n_nodes`, `n_segments`, 3) array of polynomial coefficients
            for the evaluation of the derivatives of the position (in meters)
            along the borehole segments as a function of `xi_p`.

        Returns
        -------
        float
            (3,) array of the derivative of the position along the borehole
            segment (in meters).

        """
        dp_dxi = vmap(
            jnp.polyval,
            in_axes=(1, None),
            out_axes=0
        )(dp_dxi_coefs[:, index, :], xi_p)
        return dp_dxi

    @staticmethod
    @jit
    def _longitudinal_position(xi_p: float, index: int, s_coefs: Array) -> float:
        """Longitudinal position along a borehole segment.

        Parameters
        ----------
        xi_p : float
            Coordinate along the borehole segment.
        index : int
            Index of the borehole segment.
        s_coefs : array
            (`n_nodes`, `n_segments`,) array of polynomial coefficients
            for the evaluation of the longitudinal position (in meters) along
            the borehole segments as a function of `xi_p`.

        Returns
        -------
        float
            Longitudinal position along the borehole segment (in meters).

        """
        return jnp.polyval(s_coefs[:, index], xi_p)

    @staticmethod
    @jit
    def _norm_of_jacobian(xi_p: float, index: int, J_coefs: Array) -> float:
        """Norm of the Jacobian along a borehole segment.

        Parameters
        ----------
        xi_p : float
            Coordinate along the borehole segment.
        index : int
            Index of the borehole segment.
        J_coefs : array
            (`n_nodes`, `n_segments`,) array of polynomial coefficients
            for the evaluation of the norm of the jacobian (in meters) along
            the borehole segments as a function of `xi_p`.

        Returns
        -------
        float
            Norm of the Jacobian along the borehole segment (in meters).

        """
        return jnp.polyval(J_coefs[:, index], xi_p)

    @classmethod
    @partial(jit, static_argnames=['cls'])
    def _point_heat_source(cls, xi_p: float, index: int, p: Array, time: float, alpha: float, r_min: float, p_coefs: Array) -> Array:
        """Point heat source solution.

        Parameters
        ----------
        xi_p : array
            (N,) array of coordinates along segments.
        p : array
            (3,) array of positions.
        time : float
            Time (in seconds).
        alpha : float
            Ground thermal diffusivity (in m^2/s).
        r_min : float
            Minimum distance (in meters) added to the distance between the
            heat source and the positions `p`.
        p_coefs : array
            (`n_nodes`, `n_segments`, 3) array of polynomial coefficients
            for the evaluation of the position (in meters) along
            the borehole segments as a function of `xi_p`.

        Returns
        -------
        array
            (N,) array of the point heat source solution.
        """
        p_source = cls._position(xi_p, index, p_coefs)
        p_mirror = p * jnp.array([1, 1, -1], dtype=int)
        # Point heat source solutions
        h = (
            point_heat_source(p_source, p, time, alpha, r_min)
            - point_heat_source(p_source, p_mirror, time, alpha, r_min)
        )
        return h

    @staticmethod
    @jit
    def _position(xi_p: float, index: int, p_coefs: Array) -> float:
        """Longitudinal position along a borehole segment.

        Parameters
        ----------
        xi_p : float
            Coordinate along the borehole segment.
        index : int
            Index of the borehole segment.
        p_coefs : array
            (`n_nodes`, `n_segments`, 3) array of polynomial coefficients
            for the evaluation of the position (in meters) along
            the borehole segments as a function of `xi_p`.

        Returns
        -------
        float
            (3,) array of the position along the borehole segment (in meters).

        """
        p = vmap(
            jnp.polyval,
            in_axes=(1, None),
            out_axes=0
        )(p_coefs[:, index, :], xi_p)
        return p

    @staticmethod
    @partial(jit, static_argnames=['index_out'])
    def _segment_coordinate(xi: float, xi_edges: Array, index_out: bool = True) -> Tuple[float, int] | float:
        """Segment coordinate from borehole coordinate.

        Parameters
        ----------
        xi : float
            Coordinate along the borehole.
        xi_edges : array
            (`n_segments` + 1,) array of coordinates of the edges of
            the segments.
        index_out : bool, default: True
            Set to True to return the index of the borehole segment.

        Returns
        -------
        xi_p : float
            Coordinate along the borehole segment.
        index : int, optional
            Index of the borehole segment. Returned only if `index_out`
            is True.

        """
        index = jnp.maximum(
            jnp.searchsorted(xi_edges, xi),
            1
        ).astype(int) - 1
        a, b = xi_edges[index], xi_edges[index + 1]
        xi_p = (2 * xi - (b + a)) / (b - a)
        if index_out:
            return xi_p, index
        else:
            return xi_p

    @classmethod
    @partial(jit, static_argnames=['cls', 'order'])
    def _thermal_response_factor(cls, index: int, p: Array, time: float, alpha: float, r_min: float, p_coefs: Array, J_coefs: Array, psi_coefs: Array, order: int = 101) -> Array:
        """Point heat source solution.

        Parameters
        ----------
        index : int
            Index of the borehole segment.
        p : array
            (3,) array of positions.
        time : float
            Time (in seconds).
        alpha : float
            Ground thermal diffusivity (in m^2/s).
        r_min : float
            Minimum distance (in meters) added to the distance between the
            heat source and the positions `p`.
        p_coefs : array
            (`n_nodes`, `n_segments`, 3) array of polynomial coefficients
            for the evaluation of the position (in meters) along
            the borehole segments as a function of `xi_p`.
        J_coefs : array
            (`n_nodes`, `n_segments`,) array of polynomial coefficients
            for the evaluation of the norm of the jacobian (in meters) along
            the borehole segments as a function of `xi_p`.
        psi_coefs : array
            (`n_nodes`, `n_nodes`,) array of polynomial coefficients
            for the evaluation of the polynomial basis functions along
            the borehole segments as a function of `xi_p`.
        order : int, default: 101
            Order of the numerical quadrature.

        Returns
        -------
        array
            (`n_nodes`,) array of the point heat source solution.
        """
        def thermal_reponse_factor_integrand(xi_p: float) -> Array:
            """Integrand of the thermal response factor to a point."""
            J = cls._norm_of_jacobian(xi_p, index, J_coefs)
            psi = Basis._f_psi(xi_p, psi_coefs)
            h_point_source = psi * J * cls._point_heat_source(
                xi_p, index, p, time, alpha, r_min, p_coefs)
            return h_point_source

        h = quad(
            thermal_reponse_factor_integrand,
            order=order,
            rule='gl'
        )
        return h

    @classmethod
    @partial(jit, static_argnames=['cls', 'order'])
    def _thermal_response_factor_to_self(cls, index: int, p: Array, time: float, alpha: float, r_min: float, p_coefs: Array, J_coefs: Array, psi_coefs: Array, xi: Array, order: int = 21) -> Array:
        """Point heat source solution.

        Parameters
        ----------
        index : int
            Index of the borehole segment.
        p : array
            (3,) array of positions.
        time : float
            Time (in seconds).
        alpha : float
            Ground thermal diffusivity (in m^2/s).
        r_min : float
            Minimum distance (in meters) added to the distance between the
            heat source and the positions `p`.
        p_coefs : array
            (`n_nodes`, `n_segments`, 3) array of polynomial coefficients
            for the evaluation of the position (in meters) along
            the borehole segments as a function of `xi_p`.
        J_coefs : array
            (`n_nodes`, `n_segments`,) array of polynomial coefficients
            for the evaluation of the norm of the jacobian (in meters) along
            the borehole segments as a function of `xi_p`.
        psi_coefs : array
            (`n_nodes`, `n_nodes`,) array of polynomial coefficients
            for the evaluation of the polynomial basis functions along
            the borehole segments as a function of `xi_p`.
        order : int, default: 21
            Order of the numerical quadrature along each subinterval
            delimited by segment edges and nodes.

        Returns
        -------
        array
            (`n_nodes`,) array of the point heat source solution.
        """
        def thermal_reponse_factor_to_self_integrand(xi_p: float) -> Array:
            """Integrand of the thermal response factor to a node."""
            J = cls._norm_of_jacobian(xi_p, index, J_coefs)
            psi = Basis._f_psi(xi_p, psi_coefs)
            h_point_source = psi * J * cls._point_heat_source(
                xi_p, index, p, time, alpha, r_min, p_coefs)
            return h_point_source

        points = jnp.concatenate(
            [
                -jnp.ones(1),
                xi,
                jnp.ones(1)
            ]
        )

        h = quad(
            thermal_reponse_factor_to_self_integrand,
            points=points,
            order=order,
            rule='ts'
        )
        return h

    @classmethod
    def from_dimensions(cls, L: float, D: float, r_b: float, x: float, y: float, basis: Basis, n_segments: int, tilt: float = 0., orientation: float = 0., segment_ratios: ArrayLike | None = None, order: int = 101, order_to_self: int = 21) -> Self:
        """Straight borehole from its dimensions.

        Parameters
        ----------
        L : float
            Borehole length (in meters).
        D : float
            Borehole buried depth (in meters).
        r_b : float
            Borehole radius (in meters).
        x, y : float
            Horizontal position (in meters) of the top end of the
            borehole.
        basis : basis
            Basis functions.
        n_segments : int
            Number of segments.
        tilt : float, default: ``0.``
            Tilt angle (in radians) of the borehole with respect to
            vertical.
        orientation : float, default: ``0.``
            Orientation (in radians) of the inclined borehole. An
            inclination toward the x-axis corresponds to an orientation
            of zero.
        segment_ratios : array_like or None, default: None
            Normalized size of the segments. Should total ``1``
            (i.e. ``sum(segment_ratios) = 1``). If `segment_ratios` is
            ``None``, segments of equal size are considered (i.e.
            ``segment_ratios[v] = 1 / n_segments``).
        order : int, default: 101
            Order of the Gauss-Legendre quadrature to evaluate thermal
            response factors to points outside the borehole, and to evaluate
            coefficient matrices for fluid and heat extraction rate profiles.
        order_to_self : int, default: 21
            Order of the tanh-sinh quadrature to evaluate thermal
            response factors to nodes on the borehole. Corresponds to the
            number of quadrature points along each subinterval delimited
            by nodes and edges of the segments.

        Returns
        -------
        borehole
            Instance of the `Borehole` class.

        """
        path = Path.Line(L, D, x, y, tilt, orientation)
        return cls(r_b, path, basis, n_segments, segment_ratios=segment_ratios, order=order, order_to_self=order_to_self)



# -*- coding: utf-8 -*-
from typing import Tuple

from jax import numpy as jnp
from matplotlib import pyplot as plt
from matplotlib.axes import Axes
from matplotlib.figure import Figure

from ..basis import Basis
from ._format_axis import _format_axis


def plot_basis(basis: Basis, num: int = 101, ax: Axes | None = None) -> Axes | Tuple[Figure, Axes]:
    ax_is_none = ax is None
    if ax is None:
        fig, ax = plt.subplots(layout='constrained')
    labels = [f'$\\psi_{n+1}$' for n in range(basis.n_nodes)]
    _format_axis(ax, axis_labels=[r'$\xi$', r'$\psi$'])

    xi = jnp.linspace(-1., 1., num=num)
    psi = basis.f_psi(xi)
    ax.plot(xi, psi, label=labels);
    ax.plot(basis.xi, jnp.ones(basis.n_nodes), 'k.');
    ax.legend()

    if ax_is_none:
        return fig, ax
    else:
        return ax



# -*- coding: utf-8 -*-
from collections.abc import Callable
from typing import Self, Tuple

from jax import numpy as jnp
from jax import Array, jit, vmap
import numpy as np
from scipy.special import roots_legendre


def quad(fun: Callable[[float], Array | float], points: Array = jnp.array([-1., 1.]), order: int = 51, rule: str = 'gl') -> Array | float:
    """Numerical integral.

    Parameters
    ----------
    fun : callable
        Function to integrate. Takes a float as the only argument and
        returns either an array or a float.
    points : array, default: [-1., 1.]
        Points delimiting the subintervals for the integration.
    order : int, default: 51
        Order of the numerical quadrature.
    rule : {'gl', 'ts'}, default: 'gl'
        Quadrature rule. 'gl' corresponds to Gauss-Legendre quadrature
        and 'ts' to tanh-sinh quadrature.

    Returns
    -------
    array or float
        Intregal of `fun`.

    """
    if rule.lower() == 'gl' or rule.lower() == 'gauss-legendre':
        return quadgl(fun, points=points, order=order)
    if rule.lower() == 'ts' or rule.lower() == 'tanh-sinh':
        return quadts(fun, points=points, order=order)


def quadgl(fun: Callable[[float], Array | float], points: Array = jnp.array([-1., 1.]), order: int = 51) -> Array | float:
    """Numerical integral using Gauss-Legendre rule.

    Parameters
    ----------
    fun : callable
        Function to integrate. Takes a float as the only argument and
        returns either an array or a float.
    points : array, default: [-1., 1.]
        Points delimiting the subintervals for the integration.
    order : int, default: 51
        Order of the numerical quadrature.

    Returns
    -------
    array or float
        Intregal of `fun`.

    """
    x, w = _quadgl_weights(order)
    return _quad(fun, points, x, w)


def quadts(fun: Callable[[float], Array | float], points: Array = jnp.array([-1., 1.]), order: int = 51) -> Array | float:
    """Numerical integral using tanh-sinh rule.

    Parameters
    ----------
    fun : callable
        Function to integrate. Takes a float as the only argument and
        returns either an array or a float.
    points : array, default: [-1., 1.]
        Points delimiting the subintervals for the integration.
    order : int, default: 51
        Order of the numerical quadrature.

    Returns
    -------
    array or float
        Intregal of `fun`.

    """
    x, w = _quadts_weights(order)
    return _quad(fun, points, x, w)


def _fixed_quad(fun: Callable[[float], Array | float], a: float, b: float, x: Array, w: Array) -> Array | float:
    """Numerical integral on a single interval.

    Parameters
    ----------
    fun : callable
        Function to integrate. Takes a float as the only argument and
        returns either an array or a float.
    a, b : float
        Lower and upper limit of integration.
    x : array
        Quadrature points.
    w : array
        Quadrature weights.

    Returns
    -------
    array or float
        Intregal of `fun`.

    """
    x = 0.5 * (b + a) + 0.5 * (b - a) * x
    y = vmap(fun, in_axes=0, out_axes=-1)(x)
    return 0.5 * (b - a) * y @ w


def _quad(fun: Callable[[float], Array | float], points: Array, x: Array, w: Array) -> Array | float:
    """Numerical integral over subintervals.

    Parameters
    ----------
    fun : callable
        Function to integrate. Takes a float as the only argument and
        returns either an array or a float.
    points : array
        Points delimiting the subintervals for the integration.
    x : array
        Quadrature points.
    w : array
        Quadrature weights.

    Returns
    -------
    array or float
        Intregal of `fun`.

    """
    a, b = points[:-1], points[1:]
    integral = vmap(
        lambda _a, _b: _fixed_quad(fun, _a, _b, x, w),
        in_axes=(0, 0),
        out_axes=0
    )(a, b).sum(axis=0)
    return integral


def _quadgl_weights(order: int) -> Tuple[Array, Array]:
    """Points and weights of the Gauss-Legendre quadrature.

    Parameters
    ----------
    order : int
        Order of the tanh-sinh quadrature.

    Returns
    -------
    x : array
        Quadrature points.
    w : array
        Quadrature weights.

    """
    x, w = roots_legendre(order)
    x = jnp.asarray(x)
    w = jnp.asarray(w)
    return x, w


def _quadts_weights(order: int) -> Tuple[Array, Array]:
    """Points and weights of the tanh-sinh quadrature.

    Parameters
    ----------
    order : int
        Order of the tanh-sinh quadrature.

    Returns
    -------
    x : array
        Quadrature points.
    w : array
        Quadrature weights.

    """
    one_minus_eps = jnp.array(1.0) - 10 * jnp.finfo(jnp.array(1.0).dtype).eps
    t_max = jnp.arcsinh(2. / jnp.pi * jnp.arctanh(one_minus_eps))
    t = jnp.linspace(-t_max, t_max, num=order)

    x = jnp.tanh(0.5 * jnp.pi * jnp.sinh(t))
    w = 0.5 * jnp.pi * jnp.cosh(t) / jnp.cosh(0.5 * jnp.pi * jnp.sinh(t))**2
    w = w * jnp.diff(t)[0]
    w = 2 * w / jnp.sum(w)
    return x, w



# -*- coding: utf-8 -*-
from functools import partial
from time import perf_counter
from typing import Tuple

from jax import numpy as jnp
from jax import Array, debug, jit, vmap
from jax.lax import cond, fori_loop
from jax.scipy.linalg import block_diag
from jax.typing import ArrayLike

from ..borefield.network import Network
from .load_history_reconstruction import LoadHistoryReconstruction


class gFunction:
    """g-Function.

    Parameters
    ----------
    borefield: network
        The borefield.
    m_flow : float or array
        Total fluid mass flow rate (in kg/s), or (`n_boreholes`,) array of
        fluid mass flow rate per borehole.
    time : array_like
        Times (in seconds) to evaluate the g-function.
    cp_f : float
        Fluid specific isobaric heat capacity (in J/kg-K).
    alpha : float
        Ground thermal diffusivity (in m^2/s)
    k_s : float
        Ground thermal conductivity (in W/m-K).
    p : array_like or None, default: ``None``
        (`n_points`, 3,) array of positions to evaluate the ground
        temperature. If `p` is ``None``, the ground temperature is not
        evaluated.
    disp : bool, default: ``True``
        Set to ``True`` to print initialization progression messages.

    Attributes
    ----------
    m_flow_network : float
        Total fluid mass flow rate (in kg/s).
    g : array
        The g-function of the borefield.
    q : array
        The heat extraction rate at the nodes (in W/m).
    T_b : array
        The borehole wall temperature at the nodes (in degree Celsius).
    T_f_in : array
        The inlet fluid temperature (in degree Celsius).
    T : array
        The ground temperature (in degree Celsius) at positions `p`.

    """

    def __init__(self, borefield: Network, m_flow: float | Array, cp_f: float, time: ArrayLike, alpha: float, k_s: float, p: ArrayLike | None = None, disp: bool = True):
        # Runtime type validation
        if not isinstance(time, ArrayLike):
            raise TypeError(f"Expected arraylike input; got {time}")
        if not isinstance(p, ArrayLike) and p is not None:
            raise TypeError(f"Expected arraylike or None input; got {p}")
        # Convert input to jax.Array
        time = jnp.asarray(time)
        if p is not None:
            p = jnp.asarray(p)

        # --- Class atributes ---
        # Parameters
        self.borefield = borefield
        self.n_nodes = self.borefield.n_boreholes * self.borefield.n_nodes
        self.m_flow = m_flow
        self.cp_f = cp_f
        self.time = time
        self.time = jnp.concatenate((jnp.array([0.]), self.time))
        self.alpha = alpha
        self.k_s = k_s
        self.p = p
        # Additional attributes
        self.n_times = len(time)
        if p is None:
            self.T = None
            self.n_points = 0
        else:
            self.n_points = p.shape[0]
        if len(jnp.shape(m_flow)) == 0:
            self.m_flow_network = m_flow
        else:
            self.m_flow_network = m_flow.sum()

        # --- Initialization ---
        self.loaHisRec = LoadHistoryReconstruction(
            borefield, time, alpha, p=p, disp=disp)
        self.initialize_systems_of_equations()

    def initialize_systems_of_equations(self):
        """Initialize the system of equations.

        """
        N = self.n_nodes
        # Thermal response factors
        self.h_to_self = jnp.concatenate(
            (
                jnp.zeros(
                    (1,
                     self.borefield.n_boreholes,
                     self.borefield.n_nodes,
                     self.borefield.n_boreholes,
                     self.borefield.n_nodes)
                ),
                self.loaHisRec.h_to_self / (2 * jnp.pi * self.k_s)
            ),
            axis=0)
        # Borehole heat transfer rate coefficients
        self.g_in, self.g_b = self.borefield.g_to_self(self.m_flow, self.cp_f)
        # Initialize system of equations
        self.A = jnp.block(
            [[jnp.eye(N), self.g_in.reshape((-1, 1))],
             [self.borefield.w.flatten(), jnp.zeros((1, 1))]]
            )
        self.B = jnp.zeros(N + 1)
        # Apply constant total heat extraction rate
        self.B = self.B.at[-1].set(2 * jnp.pi * self.k_s * self.borefield.L.sum())

    def update_system_of_equations(self, h_to_self: Array, T0: Array) -> Tuple[Array, Array]:
        """Update the system of equations.

        Parameters
        ----------
        h_to_self : array
            Array of thermal response factors.
        T0 : array
            Borehole wall temperature at nodes assuming zero heat
            extraction rate.

        Returns
        -------
        A : array
            Linear system of equations.
        B : array
            Right-hand side of the linear system of equations.

        """
        N = self.n_nodes
        # Update system of equations for thermal response factors at the
        # current time step `h_to_self`
        A = self.A.at[:N, :N].set(jnp.eye(N) + jnp.einsum('iml,iljn->imjn', self.g_b, h_to_self).reshape((N, N)))
        # Apply current borehole wall temperature at nodes `T0`
        B = self.B.at[:N].set(vmap(jnp.dot, in_axes=(0, 0), out_axes=0)(-self.g_b, T0.reshape((self.borefield.n_boreholes, -1))).flatten())
        return A, B

    def simulate(self, disp: bool = True, print_every: int = 10):
        """Evaluate the g-function.

        Parameters
        ----------
        disp : bool, default: ``True``
            Set to ``True`` to print simulation progression messages.
        print_every : int, default: ``10``
            Period at which simulation progression messages are printed.

        """
        tic = perf_counter()
        next_k = print_every
        if disp:
            print('Simulation start.')
        self.loaHisRec.reset_history()
        # Initialize arrays
        self.q = jnp.zeros((self.n_times, self.borefield.n_boreholes, self.borefield.n_nodes))
        self.T_b = jnp.zeros((self.n_times, self.borefield.n_boreholes, self.borefield.n_nodes))
        self.T_f_in = jnp.zeros(self.n_times)
        if self.p is not None:
            self.T = jnp.zeros((self.n_times, self.n_points))

        def simulate_step(
            i: int,
            val: Tuple[Array, Array, Array, Array, Array | None]
        ) -> Tuple[Array, Array, Array, Array, Array | None]:
            """Single simulation step."""
            # Unpack values
            time, q, T_b, T_f_in, T = val
            # Time step
            dtime = time[i + 1] - time[i]
            # Temporal and spatial superposition of past loads
            _q_reconstructed = self.loaHisRec._reconstruct_load_history(
                time[i + 1], time[1:], q)
            T0 = self.loaHisRec._temperature(
                self.loaHisRec.h_to_self,
                _q_reconstructed
            ) / (2 * jnp.pi * self.k_s)
            # Build and solve system of equations
            _q, _T_b, _T_f_in = self._simulate_step(dtime, T0)
            # Apply latest heat extraction rates
            q = self.loaHisRec._current_load(q, _q, i)
            # Store results
            T_b = T_b.at[i].set(_T_b)
            T_f_in = T_f_in.at[i].set(_T_f_in)
            # Evaluate ground temperatures
            if self.p is not None:
                _q_reconstructed = self.loaHisRec._reconstruct_load_history(
                    time[i + 1], time[1:], q)
                T = T.at[i].set(
                    self.loaHisRec._temperature_to_point(
                        self.loaHisRec.h_to_point,
                        _q_reconstructed
                    ) / (2 * jnp.pi * self.k_s)
                )
            if disp:
                toc = perf_counter()
                cond(
                    (i + 1) % print_every == 0,
                    lambda _: debug.print(
                        'Completed {i} of {n_times} time steps. ',
                        i=i+1, n_times=self.n_times
                    ),
                    lambda _: None,
                    None
                )
            return time, q, T_b, T_f_in, T
        # Pack variables
        val = self.time, self.q, self.T_b, self.T_f_in, self.T
        # Run simulation
        _, self.q, self.T_b, self.T_f_in, self.T = fori_loop(
            0, self.n_times, simulate_step, val, unroll=False)

        # Outlet fluid temperature
        T_f_out = self.T_f_in - 2 * jnp.pi * self.k_s * self.borefield.L.sum() / (self.m_flow_network * self.cp_f)
        # Average fluid temperature
        T_f = 0.5 * (self.T_f_in + T_f_out)
        # Borefield thermal resistance
        R_field = self.borefield.effective_borefield_thermal_resistance(self.m_flow, self.cp_f)
        # Effective borehole wall temperature
        self.g = T_f - 2 * jnp.pi * self.k_s * R_field
        if disp:
            toc = perf_counter()
            debug.print(
                'Simulation end. Elapsed time: {clock:.2f} seconds.',
                clock=toc-tic
            )

    @partial(jit, static_argnames=['self'])
    def _simulate_step(self, dtime: float, T0: Array) -> Tuple[Array, Array, float]:
        """Solve a single time step.

        Parameters
        ----------
        dtime : float
            Current time step variation (in seconds).
        T0 : array
            Borehole wall temperature at nodes assuming zero heat
            extraction rate.

        Returns
        -------
        q : array
            The heat extraction rate at the nodes (in W/m).
        T_b : array
            The borehole wall temperature at the nodes (in degree
            Celsius).
        T_f_in : float
            The inlet fluid temperature (in degree Celsius).

        """
        # Current thermal response factors
        h_shape = self.h_to_self.shape
        h = vmap(
            lambda _h: jnp.interp(dtime, self.time, _h),
            in_axes=-1,
            out_axes=-1
            )(
            self.h_to_self.reshape((h_shape[0], -1))
        ).reshape(h_shape[1:])
        # Build and solve system of equations
        A, B = self.update_system_of_equations(h, T0)
        X = jnp.linalg.solve(A, B)
        q = X[:self.n_nodes].reshape((self.borefield.n_boreholes, -1))
        T_b = T0 + jnp.tensordot(h, q, axes=([-2, -1], [-2, -1]))
        T_f_in = X[-1]
        return q, T_b, T_f_in
        



# -*- coding: utf-8 -*-
from collections.abc import Callable
from functools import partial
from typing import Self

from interpax import Interpolator1D
from jax import numpy as jnp
from jax import Array, jacobian, jit, vmap
from jax.typing import ArrayLike
from jax.scipy.special import erfc
from scipy.integrate import fixed_quad


class Path:
    """Trajectory of a borehole.

    Parameters
    ----------
    f_p : callable
        Function that takes a coordinate or an (N,) array of
        coordinates ``xi`` and returns an (3,) or (N, 3,) array of
        positions [x, y, z] along the path.
    f_dp_dxi : callable
        Function that takes a coordinate or an (N,) array of
        coordinates ``xi`` and returns an (3,) or (N, 3,) array of
        derivatives [dx/dxi, dy/dxi, dz/dxi] along the path.
    f_J : callable
        Function that takes a coordinate or an (N,) array of
        coordinates ``xi`` and returns a float or an (N,) array of
        norms of the Jacobian sqrt(dx/dxi**2 + dy/dxi**2 + dz/dxi**2)
        along the path.
    f_s : callable
        Function that takes a coordinate or an (N,) array of
        coordinates ``xi`` and returns a float or an (N,) array of
        the longitudinal position ``s`` (in meters) along the path.
    xi : array_like or None, default: None
        (`n_nodes`,) array of node coordinates along the interval
        ``[-1, 1]``.
    p : array_like or None, default: None
        Positions (``x``, ``y``, ``z``) of the nodes along the
        trajectory of the borehole.
    deg : int or None, default: None
        Polynomial degree of the path. This is used as a default to
        approximate the trajectory in borehole models.

    Attributes
    ----------
    n_nodes : int
        Number of nodes.
    L : float
        Length of the path (in meters)

    """

    def __init__(self, f_p: Callable[[float | Array], Array], f_dp_dxi: Callable[[float | Array], Array], f_J: Callable[[float | Array], float | Array], f_s: Callable[[float | Array], float | Array], xi: ArrayLike | None = None, p: ArrayLike | None = None, deg: int | None = None):
        # Runtime type validation
        if not isinstance(xi, ArrayLike) and xi is not None:
            raise TypeError(f"Expected arraylike input; got {xi}")
        if not isinstance(p, ArrayLike) and p is not None:
            raise TypeError(f"Expected arraylike input; got {p}")
        # Convert input to jax.Array
        xi = jnp.asarray(xi)
        p = jnp.atleast_2d(p)

        # --- Class atributes ---
        self.xi = xi
        self.p = p
        self.deg = deg

        # --- Path functions ---
        # Position (p)
        self._f_p = f_p
        # Derivative of position (dp/dxi)
        self._f_dp_dxi = f_dp_dxi
        # Norm of the Jacobian (J)
        self._f_J = f_J
        # Longitudinal position (s)
        self._f_s = f_s

        # --- Additional attributes ---
        # Number of nodes
        if xi is not None and p is not None:
            n_nodes = len(xi)
            self.n_nodes = n_nodes
        else:
            self.n_nodes = None
        # Length of the path
        self.L = self.f_s(1.)

    def f_p(self, xi: float | Array) -> Array:
        """Position along the path.

        Parameters
        ----------
        xi : float or array
            Coordinate of (N,) array of coordinates along the
            trajectory.

        Returns
        -------
        array
            (3,) or (N, 3,) array of positions
            (``x``, ``y``, ``z``) along the trajectory.
        """
        return self._f_p(xi)

    def f_dp_dxi(self, xi: float | Array) -> Array:
        """Derivative of the position along the path.

        Parameters
        ----------
        xi : float or array
            Coordinate of (N,) array of coordinates along the
            trajectory.

        Returns
        -------
        array
            (3,) or (N, 3,) array of derivatives of the position
            (``dx/dxi``, ``dy/dxi``, ``dz/xi``) along the trajectory.
        """
        return self._f_dp_dxi(xi)

    def f_J(self, xi: float | Array) -> float | Array:
        """Norm of the Jacobian along the path.

        Parameters
        ----------
        xi : float or array
            Coordinate of (N,) array of coordinates along the
            trajectory.

        Returns
        -------
        float or array
            Norm of the Jacobian or (N,) array of the norms of the
            Jacobian along the trajectory.
        """
        return self._f_J(xi)

    def f_s(self, xi: float | Array) -> float | Array:
        """Longitudinal position along the path.

        Parameters
        ----------
        xi : float or array
            Coordinate of (N,) array of coordinates along the
            trajectory.

        Returns
        -------
        float or array
            Longitudinal position or (N,) array of longitudinal
            positions along the trajectory.
        """
        return self._f_s(xi)

    @classmethod
    def Line(cls, L: float, D: float, x: float, y: float, tilt: float, orientation: float) -> Self:
        """Path from the dimensions of a borehole.

        Parameters
        ----------
        L : float
            Borehole length (in meters).
        D : float
            Borehole buried depth (in meters).
        r_b : float
            Borehole radius (in meters).
        x, y : array_like
            Horizontal position (in meters) of the top end of the
            borehole.
        tilt : float
            Tilt angle (in radians) of the borehole with respect to
            vertical.
        orientation : float
            Orientation (in radians) of the inclined borehole. An
            inclination toward the x-axis corresponds to an orientation
            of zero.

        Returns
        -------
        path
            Instance of the `Path` class.

        """
        # Position of the top of the borehole
        p0 = jnp.array([x, y, -D])
        # Position of the bottom of the borehole
        delta_p = jnp.array(
            [
                L * jnp.sin(tilt) * jnp.cos(orientation),
                L * jnp.sin(tilt) * jnp.sin(orientation),
                -L * jnp.cos(tilt)
            ]
        )
        p1 = p0 + delta_p

        # --- Class atributes ---
        xi = jnp.array([-1., 1.])
        p = jnp.stack([p0, p1], axis=0)

        # --- Path functions ---
        p_mean = p.mean(axis=0)
        L = jnp.linalg.norm(delta_p)
        def f_p(_xi: float | Array) -> Array:
            """Position along the path."""
            if len(jnp.shape(_xi)) > 0:
                return vmap(f_p, in_axes=0)(_xi)
            return p_mean + 0.5 * _xi * delta_p
        def f_dp_dxi(_xi: float | Array) -> Array:
            """Derivative of the position along the path."""
            if len(jnp.shape(_xi)) > 0:
                return jnp.broadcast_to(0.5 * delta_p, (len(_xi), 3))
            return 0.5 * delta_p
        def f_J(_xi: float | Array) -> float | Array:
            """Norm of the Jacobian along the path."""
            if len(jnp.shape(_xi)) > 0:
                return jnp.broadcast_to(0.5 * L, len(_xi))
            return 0.5 * L
        def f_s(_xi: float | Array) -> float | Array:
            """Longitudinal position along the path."""
            return 0.5 * (1 + _xi) * L
        return cls(f_p, f_dp_dxi, f_J, f_s, xi=xi, p=p, deg=1)

    @classmethod
    def Polynomial(cls, xi: ArrayLike, p: ArrayLike, deg: int = None, s_method: str = 'monotonic', s_order: int = 21, s_num: int = 21) -> Self:
        """Polynomial path from positions along the path.

        Parameters
        ----------
        xi : array_like
            (N,) array of coordinates along the path. Should be of
            sufficient length for the selected interpolation methods.
        p : array_like
            (N, 3,) array of positions (in meters) along the path.
        deg : int or None, default: None
            Degree of the polynomial representing the path. If None,
            ``deg = N - 1``. If ``deg < N - 1``, the polynomial is
            obtained by regression, taking the first position as fixed
            (usually ``xi[0] = -1.`` and its corresponding position
            ``p[0]``).
        s_method : str, default: 'monotonic'
            Interpolation methods to be used by
            ``interpax.Interpolator1D`` for the position along the
            path and for the longitudinal position ``s`` along the
            path.
        s_order : int, default: 21
            Number of points for the integration of the Jacobian between
            each subsequent coordinates to obtain the ongitudinal
            position ``s`` from the norm of the Jacobian along the path.
        s_num : int, default: 21
            Number of evenly distributed knots along the path to
            evaluate the longitudinal position ``s`` using trapezoidal
            integration and generate the interpolator.
            

        Returns
        -------
        path
            Instance of the `Path` class.

        """
        # Runtime type validation
        if not isinstance(xi, ArrayLike):
            raise TypeError(f"Expected arraylike input; got {xi}")
        if not isinstance(p, ArrayLike):
            raise TypeError(f"Expected arraylike input; got {p}")
        # Convert input to jax.Array
        xi = jnp.asarray(xi)
        p = jnp.atleast_2d(p)

        # --- Path functions ---
        # Position along the path
        if deg is None:
            p_coefs = jnp.polyfit(xi, p, len(xi)-1)
            deg = len(xi)-1
        else:
            p_w = jnp.ones_like(xi)
            p_w.at[0].set(1e4)
            p_coefs = jnp.polyfit(xi, p, deg, w=p_w)
        def f_p(_xi: float | Array) -> Array:
            """Position along the path."""
            if len(jnp.shape(_xi)) > 0:
                return vmap(f_p, in_axes=0)(_xi)
            return jnp.polyval(p_coefs, _xi)
        # Derivative of the position along the path
        dp_coefs = vmap(
            jnp.polyder,
            in_axes=-1,
            out_axes=-1
        )(p_coefs)
        def f_dp_dxi(_xi: float | Array) -> Array:
            """Derivative of the position along the path."""
            if len(jnp.shape(_xi)) > 0:
                return vmap(f_dp_dxi, in_axes=0)(_xi)
            return jnp.polyval(dp_coefs, _xi)
        # Norm of the Jacobian along the path
        dp_square_coefs = vmap(
            jnp.polymul,
            in_axes=-1,
            out_axes=-1
        )(dp_coefs, dp_coefs)
        def f_J(_xi: float | Array) -> float | Array:
            """Norm of the Jacobian along the path."""
            if len(jnp.shape(_xi)) > 0:
                return vmap(f_J, in_axes=0)(_xi)
            return jnp.sqrt(jnp.polyval(dp_square_coefs, _xi).sum(axis=-1))
        # Longitudinal position along the path
        s_xi = jnp.linspace(-1., 1., num=s_num)
        a, b = s_xi[:-1], s_xi[1:]
        ds = jnp.array(
            [
                fixed_quad(f_J, _a, _b, n=s_order)[0]
                for _a, _b in zip(a, b)
            ]
        )
        s = jnp.cumulative_sum(ds, include_initial=True)
        f_s = Interpolator1D(s_xi, s, method=s_method, extrap=True)
        return cls(f_p, f_dp_dxi, f_J, f_s, xi=xi, p=p, deg=deg)

    @classmethod
    def Spline(cls, xi: ArrayLike, p: ArrayLike, method: str = 'cubic2', s_method: str = 'monotonic', s_order: int = 21, s_num: int = 21) -> Self:
        """Path from positions along the path.

        Parameters
        ----------
        xi : array_like
            (N,) array of coordinates along the path. Should be of
            sufficient length for the selected interpolation methods.
        p : array_like
            (N, 3,) array of positions (in meters) along the path.
        method, s_method : str, default: 'cubic2', 'monotonic'
            Interpolation methods to be used by
            ``interpax.Interpolator1D`` for the position along the
            path and for the longitudinal position ``s`` along the
            path.
        s_order : int, default: 21
            Number of points for the integration of the Jacobian between
            each subsequent coordinates to obtain the ongitudinal
            position ``s`` from the norm of the Jacobian along the path.
        s_num : int, default: 21
            Number of evenly distributed knots along the path to
            evaluate the longitudinal position ``s`` using trapezoidal
            integration and generate the interpolator.
            

        Returns
        -------
        path
            Instance of the `Path` class.

        """
        # Runtime type validation
        if not isinstance(xi, ArrayLike):
            raise TypeError(f"Expected arraylike input; got {xi}")
        if not isinstance(p, ArrayLike):
            raise TypeError(f"Expected arraylike input; got {p}")
        # Convert input to jax.Array
        xi = jnp.asarray(xi)
        p = jnp.atleast_2d(p)

        # --- Path functions ---
        # Position along the path
        f_p = Interpolator1D(xi, p, method=method, extrap=True)
        # Derivative of the position along the path
        f_dp_dxi = partial(f_p, dx=1)
        # Norm of the Jacobian along the path
        def f_J(_xi: float | Array) -> float | Array:
            """Norm of the Jacobian along the path."""
            return jnp.linalg.norm(f_dp_dxi(_xi), axis=-1)
        # Longitudinal position along the path
        s_xi = jnp.linspace(-1., 1., num=s_num)
        a, b = s_xi[:-1], s_xi[1:]
        ds = jnp.array(
            [
                fixed_quad(f_J, _a, _b, n=s_order)[0]
                for _a, _b in zip(a, b)
            ]
        )
        s = jnp.cumulative_sum(ds, include_initial=True)
        f_s = Interpolator1D(s_xi, s, method=s_method, extrap=True)
        return cls(f_p, f_dp_dxi, f_J, f_s, xi=xi, p=p, deg=None)



# -*- coding: utf-8 -*-
from time import perf_counter

from jax import numpy as jnp
from jax import Array, jit, vmap
from jax.typing import ArrayLike

from ..borefield.borefield import Borefield
from ..borefield.network import Network
from ._temporal_superposition import _TemporalSuperposition


class LoadHistoryReconstruction(_TemporalSuperposition):
    """Load history reconstruction.

    Parameters
    ----------
    borefield : borefield or network
        The borefield.
    time : array
        Array of simulation times (in seconds).
    alpha : float
        Ground thermal diffusivity (in m^2/s).
    p : array_like or None, default: ``None``
        (`n_points`, 3,) array of positions to evaluate the ground
        temperature. If `p` is ``None``, the ground temperature is not
        evaluated.
    disp : bool, default: ``True``
        Set to ``True`` to print initialization progression messages.

    Attributes
    ----------
    n_times : int
        Number of time steps.
    n_points : int
        Number of points to evaluate ground temperatures.
    h_to_self : array
        (`n_times`, `n_boreholes`, `n_nodes`, `n_boreholes`,
        `n_nodes`,) array of thermal response factors at the nodes.
    h_to_point : array
        (`n_times`, `n_points`, `n_boreholes`, `n_nodes`,) array of
        thermal response factors at points.

    """

    def __init__(self, borefield: Borefield | Network, time: ArrayLike, alpha: float, p: ArrayLike | None = None, disp: bool = True):
        tic = perf_counter()
        # Runtime type validation
        if not isinstance(time, ArrayLike):
            raise TypeError(f"Expected arraylike input; got {time}")
        if not isinstance(p, ArrayLike) and p is not None:
            raise TypeError(f"Expected arraylike or None input; got {p}")
        # Convert input to jax.Array
        time = jnp.asarray(time)
        if p is not None:
            p = jnp.asarray(p)

        self.borefield = borefield
        self.time = time
        self.n_times = len(time)
        self.alpha = alpha
        self.p = p
        self.q = jnp.zeros((len(self.time), borefield.n_boreholes, borefield.n_nodes))
        self.q_reconstructed = jnp.zeros((len(self.time), borefield.n_boreholes, borefield.n_nodes))
        self._time = 0.
        self._k = -1
        if disp:
            print('Initialization start.')
        self.h_to_self = borefield.h_to_self(self.time, alpha)
        if disp:
            toc = perf_counter()
            print(
                f'Completed thermal response factors to nodes. '
                f'Elapsed time: {toc-tic:.2f} seconds.'
                    )
        if p is not None:
            self.n_points = p.shape[0]
            self.h_to_point = borefield.h_to_point(p, self.time, alpha)
            if disp:
                toc = perf_counter()
                print(
                    f'Completed thermal response factors to ground. '
                    f'Elapsed time: {toc-tic:.2f} seconds.'
                        )
        else:
            self.n_points = 0
            self.h_to_point = jnp.zeros((0, borefield.n_boreholes, borefield.n_nodes))
        if disp:
            toc = perf_counter()
            print(
                f'Initialization end. Elapsed time: {toc-tic:.2f} seconds.'
            )

    def next_time_step(self) -> float:
        """Advance to next simulation time step.

        Returns
        -------
        float
            Time (in seconds) of the new time step.

        """
        self._k += 1
        self._time = self.time[self._k]
        return self._time

    def reset_history(self):
        """Reset the history to its initial condition.

        """
        self.q = self.q.at[:].set(0.)
        self.q_reconstructed = self.q_reconstructed.at[:].set(0.)
        self._time = 0.
        self._k = -1
        return

    def set_current_load(self, q: Array):
        """Set the current heat extraction rate.

        Parameters
        ----------
        q : array
            (`n_boreholes`, `n_nodes`,) array of heat extraction rates
            (in W/m).

        """
        self.q = self._current_load(self.q, q, self._k)
        return

    def temperature(self) -> Array:
        """Evaluate temperatures at nodes.

        Returns
        ----------
        T_b : array
            (`n_boreholes`, `n_nodes`,) array of borehole wall
            temperature variations (in degree Celcius).

        """
        self.q_reconstructed = self._reconstruct_load_history(self._time, self.time, self.q)
        T = self._temperature(self.h_to_self, self.q_reconstructed)
        return T

    def temperature_to_point(self) -> Array:
        """Evaluate temperatures at point.

        Returns
        ----------
        T : array
            (`n_points`,) array of ground temperature variations
            (in degree Celcius).

        """
        self.q_reconstructed = self._reconstruct_load_history(self._time, self.time, self.q)
        T = self._temperature_to_point(self.h_to_point, self.q_reconstructed)
        return T

    @staticmethod
    def _current_load(q_history: Array, q: Array, k: int) -> Array:
        """Add current load to load history

        Parameters
        ----------
        q_history : array
            (`n_times`, `n_boreholes`, `n_nodes`,) array of loads
            (in W/m).
        q : array
            (`n_boreholes`, `n_nodes`,) array of current loads (in W/m).
        k : int
            Current time step.

        Returns
        -------
        array
            (`n_times`, `n_boreholes`, `n_nodes`,) array of updated
            loads (in W/m).

        """
        return q_history.at[k].set(q)

    @staticmethod
    @jit
    def _reconstruct_load_history(current_time: float, time: Array, q: Array) -> Array:
        """Reconstruct the load history

        Parameters
        ----------
        current_time : float
            Current time (in seconds).
        time : array
            (`n_times`,) array of time steps (in seconds).
        q : array
            (`n_times`,`n_boreholes`, `n_nodes`,) array of loads (in W/m).

        Returns
        -------
        array
            (`n_times`,`n_boreholes`, `n_nodes`,) array of reconstructed
            loads (in W/m).

        """
        time = jnp.concatenate((jnp.zeros(1), time))
        dtime = jnp.diff(time)
        q_accumulated = jnp.concatenate((jnp.zeros((1, ) + q.shape[1:]), jnp.cumsum((q.T * dtime).T, axis=0)))
        time_reconstructed = (current_time - time)[::-1]
        dtime_reconstructed = dtime[::-1]
        time_reconstructed = jnp.maximum(0., time_reconstructed)
        q_reconstructed = vmap(
            vmap(
                lambda _q: jnp.interp(time_reconstructed[1:], time, _q) - jnp.interp(time_reconstructed[:-1], time, _q),
                in_axes=1,
                out_axes=1),
            in_axes=2,
            out_axes=2
        )(q_accumulated)
        q_reconstructed = (q_reconstructed.T / dtime_reconstructed).T
        q_reconstructed = q_reconstructed.at[1:].subtract(q_reconstructed[:-1])[::-1]
        return q_reconstructed



# -*- coding: utf-8 -*-
from abc import ABC, abstractmethod
from collections.abc import Callable
from functools import partial
from typing import Self, Tuple

from jax import numpy as jnp
from jax import Array, jit, vmap
from jax.typing import ArrayLike

from ..basis import Basis
from .borehole import Borehole
from ..path import Path


class _Tube(Borehole, ABC):
    """Geothermal borehole with tube.

    Parameters
    ----------
    R_d : array_like or callable
        (`n_pipes`, `n_pipes`,) array of thermal resistances (in m-K/W),
        or callable that takes the mass flow rate as input (in kg/s) and
        returns a (`n_pipes`, `n_pipes`,) array.
    r_b : float
        Borehole radius (in meters).
    path : path
        Path of the borehole.
    basis : basis
        Basis functions.
    n_segments : int
        Number of segments.
    segment_ratios : array_like or None, default: None
        Normalized size of the segments. Should total ``1``
        (i.e. ``sum(segment_ratios) = 1``). If `segment_ratios` is
        ``None``, segments of equal size are considered (i.e.
        ``segment_ratios[v] = 1 / n_segments``).
    order : int, default: 101
        Order of the Gauss-Legendre quadrature to evaluate thermal
        response factors to points outside the borehole, and to evaluate
        coefficient matrices for fluid and heat extraction rate profiles.
    order_to_self : int, default: 21
        Order of the tanh-sinh quadrature to evaluate thermal
        response factors to nodes on the borehole. Corresponds to the
        number of quadrature points along each subinterval delimited
        by nodes and edges of the segments.
    parallel : bool, default: True
        True if pipes are in parallel. False if pipes are in series.

    Attributes
    ----------
    n_pipes : int
        Number of pipes in the borehole.
    n_nodes : int
        Total number of nodes along the borehole.
    xi_edges : array
        (`n_segments`+1,) array of the coordinates of the edges of the
        segments.
    L : float
        Borehole length (in meters).
    xi : array
        (`n_nodes`,) array of node coordinates.
    p : array
        (`n_nodes`, 3,) array of node positions.
    dp_dxi : array
        (`n_nodes`, 3,) array of the derivatives of the position at the
        node coordinates.
    J : array
        (`n_nodes`,) array of the norm of the Jacobian at the node
        coordinates.
    s : array
        (`n_nodes`,) array of the longitudinal position at the node
        coordinates.
    w : array
        (`n_nodes`,) array of quadrature weights at the node coordinates.
        These quadrature weights take into account the norm of the
        Jacobian.

    """

    def __init__(self, R_d: ArrayLike | Callable[[float], Array], r_b: float, path: Path, basis: Basis, n_segments: int, segment_ratios: ArrayLike | None = None, order: int = 101, order_to_self: int = 21, parallel: bool = True):
        # Runtime type validation
        if not isinstance(R_d, ArrayLike) and not callable(R_d):
            raise TypeError(f"Expected arraylike or callable input; got {R_d}")
        # Convert input to jax.Array
        if isinstance(R_d, ArrayLike):
            R_d = jnp.atleast_2d(R_d)
            def thermal_resistances(m_flow: float) -> Array:
                return R_d
            self.thermal_resistances = thermal_resistances
        else:
            self.thermal_resistances = R_d

        # Initialize borehole
        super().__init__(r_b, path, basis, n_segments, segment_ratios=segment_ratios)
        # Other attributes
        self.R_d = R_d
        self.parallel = parallel
        shape = jnp.shape(self.thermal_resistances(1.))
        self.n_pipes = shape[0]
        self._n_pipes_over_two = int(self.n_pipes / 2)
        # Coefficients related to pipe configuration
        if self.parallel:
            # Factor for flow division amongst pipes
            self._m_flow_factor = 1 / self._n_pipes_over_two
            # Coefficients for connectivity at the top of the borehole
            # T_f_d(-1) = c_in * T_f_in + c_u @ T_f_u(-1)
            _top_connectivity_in = jnp.ones(self._n_pipes_over_two)
            _top_connectivity_u = jnp.zeros(
                (self._n_pipes_over_two, self._n_pipes_over_two)
            )
            # Coefficient for mixing of the fluid at the outlet
            # T_f_out = m_u @ T_f_u(-1)
            self._mixing = jnp.full(
                self._n_pipes_over_two,
                self._m_flow_factor
            )
        else:
            # Factor for flow division amongst pipes
            self._m_flow_factor = 1
            # Coefficients for connectivity at the top of the borehole
            # T_f_d(-1) = c_in * T_f_in + c_u @ T_f_u(-1)
            _top_connectivity_in = jnp.concatenate([
                jnp.ones(1),
                jnp.zeros(self._n_pipes_over_two - 1),
            ])
            _top_connectivity_u = jnp.eye(
                self._n_pipes_over_two,
                k=-1
            )
            # Coefficient for mixing of the fluid at the outlet
            # T_f_out = m_u @ T_f_u(-1)
            self._mixing = jnp.concatenate([
                jnp.zeros(self._n_pipes_over_two - 1),
                jnp.ones(1),
            ])
        self._top_connectivity = (_top_connectivity_in, _top_connectivity_u)

    def effective_borehole_thermal_resistance(self, m_flow: float, cp_f: float) -> float:
        """Effective borehole thermal resistance.

        Parameters
        ----------
        m_flow : float
            Fluid mass flow rate (in kg/s).
        cp_f : float
            Fluid specific isobaric heat capacity (in J/kg-K).

        Returns
        -------
        float
            Effective borehole thermal resistance (in m-K/W).

        """
        m_flow_pipe = self.m_flow_pipe(m_flow)
        beta_ij = self._beta_ij(m_flow, cp_f)
        a = self._outlet_fluid_temperature_a_in(
            beta_ij,
            self._top_connectivity,
            self._mixing,
            self._s_coefs)
        R_b = 0.5 * self.L / (m_flow * cp_f) * (1. + a) / (1. - a)
        return R_b

    def g(self, xi: Array | float, m_flow: float, cp_f: float) -> Tuple[Array | float, Array]:
        """Coefficients to evaluate the heat extraction rate.

        Parameters
        ----------
        xi : array or float
            (M,) array of coordinates along the borehole.
        m_flow : float
            Fluid mass flow rate (in kg/s).
        cp_f : float
            Fluid specific isobaric heat capacity (in J/kg-K).

        Returns
        -------
        a_in : array or float
            (M,) array of coefficients for the inlet fluid temperature.
        a_b : array
            (M, `n_nodes`,) array of coefficients for the borehole wall
            temperature.

        """
        a_in, a_b = self._heat_extraction_rate(xi, m_flow, cp_f)
        return a_in, a_b

    def g_to_self(self, m_flow: float, cp_f: float) -> Tuple[Array, Array]:
        """Coefficients to evaluate the heat extraction rate at nodes.

        Parameters
        ----------
        m_flow : float
            Fluid mass flow rate (in kg/s).
        cp_f : float
            Fluid specific isobaric heat capacity (in J/kg-K).

        Returns
        -------
        a_in : array
            (`n_nodes`,) array of coefficients for the inlet fluid
            temperature.
        a_b : array
            (`n_nodes`, `n_nodes`,) array of coefficients for the borehole
            wall temperature.

        """
        a_in, a_b = self._heat_extraction_rate_to_self(m_flow, cp_f)
        return a_in, a_b

    def fluid_temperature(self, xi: Array | float, T_f_in: float, T_b: Array, m_flow: float, cp_f: float) -> Array:
        """Fluid temperatures.

        Parameters
        ----------
        xi : array or float
            (M,) array of coordinates along the borehole.
        T_f_in : float
            Inlet fluid temperature (in degree Celsius).
        T_b : array
            (`n_nodes`,) array of borehole wall temperatures (in degree
            Celsius).
        m_flow : float
            Fluid mass flow rate (in kg/s).
        cp_f : float
            Fluid specific isobaric heat capacity (in J/kg-K).

        Returns
        -------
        array
            (M, 2,) array of fluid temperatures (in degree Celsius).

        """
        beta_ij = self._beta_ij(m_flow, cp_f)
        a_in, a_b = self._fluid_temperature(xi, m_flow, cp_f)
        T_f = a_in * T_f_in + a_b @ T_b
        return T_f

    def heat_extraction_rate(self, xi: Array | float, T_f_in: float, T_b: Array, m_flow: float, cp_f: float) -> Array | float:
        """Heat extraction rate.

        Parameters
        ----------
        xi : array or float
            (M,) array of coordinates along the borehole.
        T_f_in : float
            Inlet fluid temperature (in degree Celsius).
        T_b : array
            (`n_nodes`,) array of borehole wall temperatures (in degree
            Celsius).
        m_flow : float
            Fluid mass flow rate (in kg/s).
        cp_f : float
            Fluid specific isobaric heat capacity (in J/kg-K).

        Returns
        -------
        array
            (M,) array of heat extraction rate (in W/m). If `xi` is a
            ``float``, then ``M=0``.

        """
        a_in, a_b = self._heat_extraction_rate(xi, m_flow, cp_f)
        q = a_in * T_f_in + a_b @ T_b
        return q

    def heat_extraction_rate_to_self(self, T_f_in: float, T_b: Array, m_flow: float, cp_f: float) -> Array:
        """Heat extraction rate at nodes.

        Parameters
        ----------
        T_f_in : float
            Inlet fluid temperature (in degree Celsius).
        T_b : array
            (`n_nodes`,) array of borehole wall temperatures (in degree
            Celsius).
        m_flow : float
            Fluid mass flow rate (in kg/s).
        cp_f : float
            Fluid specific isobaric heat capacity (in J/kg-K).

        Returns
        -------
        array
            (`n_nodes`,) array of heat extraction rate (in W/m). If `xi`
            is a ``float``, then ``M=0``.

        """
        a_in, a_b = self._heat_extraction_rate_to_self(m_flow, cp_f)
        q = a_in * T_f_in + a_b @ T_b
        return q

    def m_flow_pipe(self, m_flow: float) -> Array:
        """Fluid mass flow rate per pipe.

        Parameters
        ----------
        m_flow : float
            Fluid mass flow rate into the borehole (in kg/s).

        Returns
        -------
        float
            Fluid mass flow rate in each pipes (in kg/s).

        """
        return self._m_flow_factor * m_flow

    def outlet_fluid_temperature(self, T_f_in: float, T_b: Array, m_flow: float, cp_f: float) -> float:
        """Outlet fluid temperature.

        Parameters
        ----------
        T_f_in : float
            Inlet fluid temperature (in degree Celsius).
        T_b : array
            (`n_nodes`,) array of borehole wall temperatures (in degree
            Celsius).
        m_flow : float
            Fluid mass flow rate (in kg/s).
        cp_f : float
            Fluid specific isobaric heat capacity (in J/kg-K).

        Returns
        -------
        float
            Outlet fluid temperature (in degree Celsius).

        """
        a_in, a_b = self._outlet_fluid_temperature(m_flow, cp_f)
        T_f_out = a_in * T_f_in + a_b @ T_b
        return T_f_out

    def _beta_ij(self, m_flow: float, cp_f: float) -> Array:
        """Thermal conductance coefficients.

        Parameters
        ----------
        m_flow : float
            Fluid mass flow rate (in kg/s).
        cp_f : float
            Fluid specific isobaric heat capacity (in J/kg-K).

        Returns
        -------
        array
            (`n_pipes`, `n_pipes`,) array of thermal conductance
            coefficients.

        """
        R_d = self.thermal_resistances(m_flow)
        return 1. / (self.m_flow_pipe(m_flow) * cp_f * R_d)

    def _fluid_temperature(self, xi: Array | float, m_flow: float, cp_f: float) -> Tuple[Array, Array]:
        """Coefficients to evaluate the fluid temperatures.

        Parameters
        ----------
        xi : array or float
            (M,) array of coordinates along the borehole.
        m_flow : float
            Fluid mass flow rate (in kg/s).
        cp_f : float
            Fluid specific isobaric heat capacity (in J/kg-K).

        Returns
        -------
        a_in : array
            (M, `n_pipes`,) array of coefficients for the inlet fluid
            temperature.
        a_b : array
            (M, `n_pipes`, `n_nodes`,) array of coefficients for the
            borehole wall temperature.

        """
        beta_ij = self._beta_ij(m_flow, cp_f)
        if len(jnp.shape(xi)) > 0:
            xi_p, index = vmap(
                self._segment_coordinate,
                in_axes=(0, None),
                out_axes=0
            )(xi, self.xi_edges)
            a_in = vmap(
                self._fluid_temperature_a_in,
                in_axes=(0, 0, None, None, None),
                out_axes=0
            )(xi_p, index, beta_ij, self._top_connectivity, self._s_coefs)
            a_b = vmap(
                self._fluid_temperature_a_b,
                in_axes=(0, 0, None, None, None, None, None, None),
                out_axes=0
            )(
                xi_p,
                index,
                beta_ij,
                self._top_connectivity,
                self._s_coefs,
                self._J_coefs,
                self.basis._psi_coefs,
                self.order
            )
        else:
            xi_p, index = self._segment_coordinate(xi, self.xi_edges)
            a_in = self._fluid_temperature_a_in(
                xi_p,
                index,
                beta_ij,
                self._top_connectivity,
                self._s_coefs)
            a_b = self._fluid_temperature_a_b(
                xi_p,
                index,
                beta_ij,
                self._top_connectivity,
                self._s_coefs,
                self._J_coefs,
                self.basis._psi_coefs,
                self.order
            )
        return a_in, a_b

    def _heat_extraction_rate(self, xi: Array | float, m_flow: float, cp_f: float) -> Tuple[Array | float, Array]:
        """Coefficients to evaluate the heat extraction rate.

        Parameters
        ----------
        xi : array or float
            (M,) array of coordinates along the borehole.
        m_flow : float
            Fluid mass flow rate (in kg/s).
        cp_f : float
            Fluid specific isobaric heat capacity (in J/kg-K).

        Returns
        -------
        a_in : array or float
            (M,) array of coefficients for the inlet fluid temperature.
        a_b : array
            (M, `n_nodes`,) array of coefficients for the borehole wall
            temperature.

        """
        m_flow_pipe = self.m_flow_pipe(m_flow)
        beta_ij = self._beta_ij(m_flow, cp_f)
        if len(jnp.shape(xi)) > 0:
            xi_p, index = vmap(
                self._segment_coordinate,
                in_axes=(0, None),
                out_axes=0
            )(xi, self.xi_edges)
            a_in = vmap(
                self._heat_extraction_rate_a_in,
                in_axes=(0, 0, None, None, None, None, None),
                out_axes=0
            )(
                xi_p,
                index,
                m_flow_pipe,
                cp_f,
                beta_ij,
                self._top_connectivity,
                self._s_coefs
            )
            a_b = vmap(
                self._heat_extraction_rate_a_b,
                in_axes=(0, 0, None, None, None, None, None, None, None, None),
                out_axes=0
            )(
                xi_p,
                index,
                m_flow_pipe,
                cp_f,
                beta_ij,
                self._top_connectivity,
                self._s_coefs,
                self._J_coefs,
                self.basis._psi_coefs,
                self.order
            )
        else:
            xi_p, index = self._segment_coordinate(xi, self.xi_edges)
            a_in = self._heat_extraction_rate_a_in(
                xi_p,
                index,
                m_flow_pipe,
                cp_f,
                beta_ij,
                self._top_connectivity,
                self._s_coefs
            )
            a_b = self._heat_extraction_rate_a_b(
                xi_p,
                index,
                m_flow_pipe,
                cp_f,
                beta_ij,
                self._top_connectivity,
                self._s_coefs,
                self._J_coefs,
                self.basis._psi_coefs,
                self.order
            )
        return a_in, a_b

    def _heat_extraction_rate_to_self(self, m_flow: float, cp_f: float) -> Tuple[Array | float, Array]:
        """Coefficients to evaluate the heat extraction rate.

        Parameters
        ----------
        m_flow : float
            Fluid mass flow rate (in kg/s).
        cp_f : float
            Fluid specific isobaric heat capacity (in J/kg-K).

        Returns
        -------
        a_in : array or float
            (`n_nodes`,) array of coefficients for the inlet fluid temperature.
        a_b : array
            (`n_nodes`, `n_nodes`,) array of coefficients for the borehole wall
            temperature.

        """
        m_flow_pipe = self.m_flow_pipe(m_flow)
        beta_ij = self._beta_ij(m_flow, cp_f)
        xi_p = self.basis.xi
        index = jnp.arange(self.n_segments)
        a_in = vmap(
            vmap(
                self._heat_extraction_rate_a_in,
                in_axes=(0, None, None, None, None, None, None),
                out_axes=0),
            in_axes=(None, 0, None, None, None, None, None),
            out_axes=0
        )(
            xi_p,
            index,
            m_flow_pipe,
            cp_f,
            beta_ij,
            self._top_connectivity,
            self._s_coefs
        )
        a_b = vmap(
            vmap(
                self._heat_extraction_rate_a_b,
                in_axes=(0, None, None, None, None, None, None, None, None, None),
                out_axes=0),
            in_axes=(None, 0, None, None, None, None, None, None, None, None),
            out_axes=0
        )(
            xi_p,
            index,
            m_flow_pipe,
            cp_f,
            beta_ij,
            self._top_connectivity,
            self._s_coefs,
            self._J_coefs,
            self.basis._psi_coefs,
            self.order
        )
        return a_in.flatten(), a_b.reshape(self.n_nodes, self.n_nodes)

    def _outlet_fluid_temperature(self, m_flow: float, cp_f: float) -> Tuple[float, Array]:
        """Coefficients to evaluate the outlet fluid temperature.

        Parameters
        ----------
        m_flow : float
            Fluid mass flow rate (in kg/s).
        cp_f : float
            Fluid specific isobaric heat capacity (in J/kg-K).

        Returns
        -------
        a_in : float
            Coefficient for the inlet fluid temperature.
        a_b : array
            (`n_nodes`,) array of coefficients for the borehole wall
            temperature.

        """
        beta_ij = self._beta_ij(m_flow, cp_f)
        a_in = self._outlet_fluid_temperature_a_in(
            beta_ij,
            self._top_connectivity,
            self._mixing,
            self._s_coefs
        )
        a_b = self._outlet_fluid_temperature_a_b(
            beta_ij,
            self._top_connectivity,
            self._mixing,
            self._s_coefs,
            self._J_coefs,
            self.basis._psi_coefs,
            self.order
        )
        return a_in, a_b

    @classmethod
    @abstractmethod
    def _fluid_temperature_a_in(cls, xi_p: float, index: int, beta_ij: Array, top_connectivity: Tuple[Array, Array], s_coefs: Array) -> Array:
        """Inlet coefficient to evaluate the fluid temperatures.

        Parameters
        ----------
        xi_p : float
            Coordinate along the borehole segment.
        index : int
            Index of the borehole segment.
        beta_ij : array
            (`n_pipes`, `n_pipes`,) array of thermal conductance
            coefficients.
        top_connectivity : tuple of array
            Tuple of two arrays (``c_in`` and ``c_u``) of shape
            (`n_pipes`/2,) and (`n_pipes`/2, `n_pipes`/2,). The two
            arrays give the relation between the inlet fluid temperature,
            the fluid temperatures at the top-end of the borehole in the
            upward flowing pipes, and the fluid temperatures at the top-end
            of the borehole in the downward flowing pipes following the
            relation: ``T_fd = c_in * T_f_in + c_u @ T_fu``.
        s_coefs : array
            (`n_nodes`, `n_segments`,) array of polynomial coefficients
            for the evaluation of the longitudinal position (in meters) along
            the borehole as a function of `xi_p`.

        Returns
        -------
        array
            (`n_pipes`,) array of coefficients for the inlet fluid
            temperature.

        """
        ...

    @classmethod
    @abstractmethod
    def _fluid_temperature_a_b(cls, xi_p: float, index: int, beta_ij: Array, top_connectivity: Tuple[Array, Array], s_coefs: Array, J_coefs: Array, psi_coefs: Array, order: int = 101) -> Array:
        """Borehole wall coefficient to evaluate the fluid temperatures.

        Parameters
        ----------
        xi_p : float
            Coordinate along the borehole segment.
        index : int
            Index of the borehole segment.
        beta_ij : array
            (`n_pipes`, `n_pipes`,) array of thermal conductance
            coefficients.
        top_connectivity : tuple of array
            Tuple of two arrays (``c_in`` and ``c_u``) of shape
            (`n_pipes`/2,) and (`n_pipes`/2, `n_pipes`/2,). The two
            arrays give the relation between the inlet fluid temperature,
            the fluid temperatures at the top-end of the borehole in the
            upward flowing pipes, and the fluid temperatures at the top-end
            of the borehole in the downward flowing pipes following the
            relation: ``T_fd = c_in * T_f_in + c_u @ T_fu``.
        s_coefs : array
            (`n_nodes`, `n_segments`,) array of polynomial coefficients
            for the evaluation of the longitudinal position (in meters) along
            the borehole segments as a function of `xi_p`.
        J_coefs : array
            (`n_nodes`, `n_segments`,) array of polynomial coefficients
            for the evaluation of the norm of the jacobian (in meters) along
            the borehole segments as a function of `xi_p`.
        psi_coefs : array
            (`n_nodes`, `n_nodes`,) array of polynomial coefficients
            for the evaluation of the polynomial basis functions along
            the borehole segments as a function of `xi_p`.
        order : int, default: 101
            Order of the numerical quadrature.

        Returns
        -------
        array
            (`n_pipes`, `n_nodes`,) array of coefficients for the borehole wall
            temperature.

        """
        ...

    @classmethod
    @partial(jit, static_argnames=['cls'])
    def _heat_extraction_rate_a_in(cls, xi_p: float, index: int, m_flow_pipe: float, cp_f: float, beta_ij: Array, top_connectivity: Tuple[Array, Array], s_coefs: Array) -> float:
        """Inlet coefficient to evaluate the heat extraction rate.

        Parameters
        ----------
        xi_p : float
            Coordinate along the borehole segment.
        index : int
            Index of the borehole segment.
        m_flow_pipe : float
            Fluid mass flow rate in the pipes (in kg/s).
        cp_f : float
            Fluid specific isobaric heat capacity (in J/kg-K).
        beta_ij: array
            (`n_pipes`, `n_pipes`,) array of thermal conductance
            coefficients.
        top_connectivity : tuple of array
            Tuple of two arrays (``c_in`` and ``c_u``) of shape
            (`n_pipes`/2,) and (`n_pipes`/2, `n_pipes`/2,). The two
            arrays give the relation between the inlet fluid temperature,
            the fluid temperatures at the top-end of the borehole in the
            upward flowing pipes, and the fluid temperatures at the top-end
            of the borehole in the downward flowing pipes following the
            relation: ``T_fd = c_in * T_f_in + c_u @ T_fu``.
        s_coefs : array
            (`n_nodes`, `n_segments`,) array of polynomial coefficients
            for the evaluation of the longitudinal position (in meters) along
            the borehole segments as a function of `xi_p`.

        Returns
        -------
        float
            Coefficient for the inlet fluid temperature.

        """
        b_in = cls._fluid_temperature_a_in(
            xi_p,
            index,
            beta_ij,
            top_connectivity,
            s_coefs
        )
        R_d = 1 / (m_flow_pipe * cp_f * beta_ij)
        R_d_diag = jnp.diag(R_d)
        G_d_diag = 1 / R_d_diag
        a_in = -G_d_diag @ b_in
        return a_in

    @classmethod
    @partial(jit, static_argnames=['cls', 'order'])
    def _heat_extraction_rate_a_b(cls, xi_p: float, index: int, m_flow_pipe: float, cp_f: float, beta_ij: Array, top_connectivity: Tuple[Array, Array], s_coefs: Array, J_coefs: Array, psi_coefs: Array, order: int = 101) -> Array:
        """Borehole wall coefficient to evaluate the heat extraction rate.

        Parameters
        ----------
        xi_p : float
            Coordinate along the borehole segment.
        index : int
            Index of the borehole segment.
        m_flow_pipe : float
            Fluid mass flow rate (in kg/s).
        cp_f : float
            Fluid specific isobaric heat capacity (in J/kg-K).
        beta_ij: array
            (`n_pipes`, `n_pipes`,) array of thermal conductance coefficients.
        top_connectivity : tuple of array
            Tuple of two arrays (``c_in`` and ``c_u``) of shape
            (`n_pipes`/2,) and (`n_pipes`/2, `n_pipes`/2,). The two
            arrays give the relation between the inlet fluid temperature,
            the fluid temperatures at the top-end of the borehole in the
            upward flowing pipes, and the fluid temperatures at the top-end
            of the borehole in the downward flowing pipes following the
            relation: ``T_fd = c_in * T_f_in + c_u @ T_fu``.
        s_coefs : array
            (`n_nodes`, `n_segments`,) array of polynomial coefficients
            for the evaluation of the longitudinal position (in meters) along
            the borehole segments as a function of `xi_p`.
        J_coefs : array
            (`n_nodes`, `n_segments`,) array of polynomial coefficients
            for the evaluation of the norm of the jacobian (in meters) along
            the borehole segments as a function of `xi_p`.
        psi_coefs : array
            (`n_nodes`, `n_nodes`,) array of polynomial coefficients
            for the evaluation of the polynomial basis functions along
            the borehole segments as a function of `xi_p`.
        order : int, default: 101
            Order of the numerical quadrature.

        Returns
        -------
        array
            (`n_nodes`,) array of coefficients for the borehole wall
            temperature.

        """
        a_b = jnp.zeros((jnp.shape(s_coefs)[1], jnp.shape(psi_coefs)[0]))
        b_b = cls._fluid_temperature_a_b(
            xi_p,
            index,
            beta_ij,
            top_connectivity,
            s_coefs,
            J_coefs,
            psi_coefs,
            order
        )
        R_d = 1 / (m_flow_pipe * cp_f * beta_ij)
        R_d_diag = jnp.diag(R_d)
        G_d_diag = 1 / R_d_diag
        R_b = 1 / (1 / R_d_diag).sum()
        psi = Basis._f_psi(xi_p, psi_coefs)
        a_b = a_b.at[index, :].set(psi * G_d_diag.sum()).flatten()
        a_b = a_b.at[:].add(-G_d_diag @ b_b)
        return a_b

    @classmethod
    @abstractmethod
    def _outlet_fluid_temperature_a_in(cls, beta_ij: Array, top_connectivity: Tuple[Array, Array], mixing: Array, s_coefs: Array) -> float:
        """Inlet coefficient to evaluate the outlet fluid temperature.

        Parameters
        ----------
        beta_ij: array
            (`n_pipes`, `n_pipes`,) array of thermal conductance
            coefficients.
        top_connectivity : tuple of array
            Tuple of two arrays (``c_in`` and ``c_u``) of shape
            (`n_pipes`/2,) and (`n_pipes`/2, `n_pipes`/2,). The two
            arrays give the relation between the inlet fluid temperature,
            the fluid temperatures at the top-end of the borehole in the
            upward flowing pipes, and the fluid temperatures at the top-end
            of the borehole in the downward flowing pipes following the
            relation: ``T_fd = c_in * T_f_in + c_u @ T_fu``.
        mixing : array
            (`n_pipes`/2,) array of coefficients to evaluate the outlet fluid
            temperature from the fluid temperatures at the top-end of the
            borehole in the upward flowing pipes following the relation:
            ``T_f_out = mixing @ T_fu``.
        s_coefs : array
            (`n_nodes`, `n_segments`,) array of polynomial coefficients
            for the evaluation of the longitudinal position (in meters) along
            the borehole segments as a function of `xi_p`.

        Returns
        -------
        float
            Coefficient for the inlet fluid temperature.

        """
        ...

    @classmethod
    @abstractmethod
    def _outlet_fluid_temperature_a_b(cls, beta_ij: Array, top_connectivity: Tuple[Array, Array], mixing: Array, s_coefs: Array, J_coefs: Array, psi_coefs: Array, order: int = 101) -> Array:
        """Borehole coefficient to evaluate the outlet fluid temperature.

        Parameters
        ----------
        beta_ij: array
            (`n_pipes`, `n_pipes`,) array of thermal conductance
            coefficients.
        top_connectivity : tuple of array
            Tuple of two arrays (``c_in`` and ``c_u``) of shape
            (`n_pipes`/2,) and (`n_pipes`/2, `n_pipes`/2,). The two
            arrays give the relation between the inlet fluid temperature,
            the fluid temperatures at the top-end of the borehole in the
            upward flowing pipes, and the fluid temperatures at the top-end
            of the borehole in the downward flowing pipes following the
            relation: ``T_fd = c_in * T_f_in + c_u @ T_fu``.
        mixing : array
            (`n_pipes`/2,) array of coefficients to evaluate the outlet fluid
            temperature from the fluid temperatures at the top-end of the
            borehole in the upward flowing pipes following the relation:
            ``T_f_out = mixing @ T_fu``.
        s_coefs : array
            (`n_nodes`, `n_segments`,) array of polynomial coefficients
            for the evaluation of the longitudinal position (in meters) along
            the borehole segments as a function of `xi_p`.
        J_coefs : array
            (`n_nodes`, `n_segments`,) array of polynomial coefficients
            for the evaluation of the norm of the jacobian (in meters) along
            the borehole segments as a function of `xi_p`.
        psi_coefs : array
            (`n_nodes`, `n_nodes`,) array of polynomial coefficients
            for the evaluation of the polynomial basis functions along
            the borehole segments as a function of `xi_p`.
        order : int, default: 101
            Order of the numerical quadrature.

        Returns
        -------
        array
            (`n_nodes`,) array of coefficients for the borehole wall
            temperature.

        """
        ...

    @classmethod
    def from_dimensions(cls, R_d: ArrayLike | Callable[[float], Array], L: float, D: float, r_b: float, x: float, y: float, basis: Basis, n_segments: int, tilt: float = 0., orientation: float = 0., segment_ratios: ArrayLike | None = None, order: int = 101, order_to_self: int = 21, parallel: bool = True) -> Self:
        """Straight borehole from its dimensions.

        Parameters
        ----------
        R_d : array_like or callable
            (`n_pipes`, `n_pipes`,) array of thermal resistances
            (in m-K/W), or callable that takes the mass flow rate as input
            (in kg/s) and returns a (`n_pipes`, `n_pipes`,) array.
        L : float
            Borehole length (in meters).
        D : float
            Borehole buried depth (in meters).
        r_b : float
            Borehole radius (in meters).
        x, y : float
            Horizontal position (in meters) of the top end of the
            borehole.
        basis : basis
            Basis functions.
        n_segments : int
            Number of segments.
        tilt : float, default: ``0.``
            Tilt angle (in radians) of the borehole with respect to
            vertical.
        orientation : float, default: ``0.``
            Orientation (in radians) of the inclined borehole. An
            inclination toward the x-axis corresponds to an orientation
            of zero.
        segment_ratios : array_like or None, default: None
            Normalized size of the segments. Should total ``1``
            (i.e. ``sum(segment_ratios) = 1``). If `segment_ratios` is
            ``None``, segments of equal size are considered (i.e.
            ``segment_ratios[v] = 1 / n_segments``).
        order : int, default: 101
            Order of the Gauss-Legendre quadrature to evaluate thermal
            response factors to points outside the borehole, and to evaluate
            coefficient matrices for fluid and heat extraction rate profiles.
        order_to_self : int, default: 21
            Order of the tanh-sinh quadrature to evaluate thermal
            response factors to nodes on the borehole. Corresponds to the
            number of quadrature points along each subinterval delimited
            by nodes and edges of the segments.
        parallel : bool, default: True
            True if pipes are in parallel. False if pipes are in series.

        Returns
        -------
        borehole
            Instance of the `Borehole` class.

        """
        path = Path.Line(L, D, x, y, tilt, orientation)
        return cls(R_d, r_b, path, basis, n_segments, segment_ratios=segment_ratios, order=order, order_to_self=order_to_self, parallel=parallel)
